////////////////////////////////////////////////////////////////////////////////
// Copyright (c) Elektrobit Automotive GmbH
// Alle Rechte vorbehalten. All Rights Reserved.
//
// Information contained herein is subject to change without notice.
// Elektrobit retains ownership and all other rights in the software and each
// component thereof.
// Any reproduction of the software or components thereof without the prior
// written permission of Elektrobit is prohibited.
////////////////////////////////////////////////////////////////////////////////

#include "DirectX11RippleEffect.h"
#include <gtf/scdr/sdk/GraphicsExecutionContext.h>
#include <gtf/scdr/sdk/GraphicsHandles.h>
#include <gtf/scdr/sdk/Color.h>

#include <d3dcompiler.h>
#include <stdio.h>


static const char* VertexShader =
"cbuffer cbViewPojection : register(b0)                           "
"{                                                                "
"    float4x4 projMat;                                            "
"    float4x4 viewMat;                                            "
"};                                                               "
"struct VS_IN                                                     "
"{                                                                "
"    float2 pos : POSITION0;                                      "
"    float2 tex : TEXCOORD0;                                      "
"};                                                               "
"struct VS_OUT                                                    "
"{                                                                "
"    float4 pos : SV_POSITION;                                    "
"    float2 tex : TEXCOORD0;                                      "
"};                                                               "
"VS_OUT main(VS_IN vsIn)                                          "
"{                                                                "
"    VS_OUT vsOut;                                                "
"    float4 view = mul(float4(vsIn.pos, 0, 1), viewMat);          "
"    vsOut.pos = mul(view, projMat);                              "
"    vsOut.tex = vsIn.tex;                                        "
"    return vsOut;                                                "
"}                                                                ";

static const char* PixelShader =
"Texture2D<float4> tex : register(t0);                                              "
"SamplerState texSampler : register(s0);                                            "
"cbuffer cbRipple : register(b0)                                                    "
"{                                                                                  "
"    float time;                                                                    "
"    float active;                                                                  "
"}                                                                                  "
"struct PS_IN                                                                       "
"{                                                                                  "
"    float4 pos : SV_POSITION;                                                      "
"    float2 tex : TEXCOORD0;                                                        "
"};                                                                                 "
"struct PS_OUT                                                                      "
"{                                                                                  "
"    float4 color : SV_TARGET0;                                                     "
"};                                                                                 "
"PS_OUT main(PS_IN psIn)                                                            "
"{                                                                                  "
"    PS_OUT psOut;                                                                  "
"    float2 p = -1.f + 2.f * psIn.tex;                                              "
"    float l = length(p);                                                           "
"    float2 uv = psIn.tex + active * (p / l) * cos(l * 12 - 4 * time / 100) * 0.03; "
"    float4 texColor = tex.SampleLevel(texSampler, uv, 0);                          "
"    psOut.color = texColor;                                                        "
"    return psOut;                                                                  "
"}                                                                                  ";

uint16_t DirectX11RippleEffect::propertyHandles[] = { 0, 0 };

// OpenGL|ES 2.0 and GTF expects column major matrices, DirectX11 expects
// row major matrices. This helper function copies a matrix and converts 
// it for the DirectX11 API.
static void SetConstantBufferMatrix(const gtf::math::float4x4& matrixIn, FLOAT* cbMatrixOut)
{
    const float* values = matrixIn.values;
    for (uint32_t x = 0; x < 4; ++x)
    {
        for (uint32_t y = 0; y < 4; ++y)
        {
            const uint32_t srcIdx = x * 4 + y;
            const uint32_t dstIdx = y * 4 + x;
            cbMatrixOut[dstIdx] = static_cast<FLOAT>(values[srcIdx]);
        }
    }
}

// Callback for instantiating a new renderer. This callback is set in 
// RippleEffectPlugin::registerRenderer
gtf::scdr::sdk::WidgetRenderer* DirectX11RippleEffect::createIntance(
    const gtf::scdr::sdk::WidgetRendererRegistration& registration_,
    gtf::scdr::sdk::GraphicsContext& context_,
    void*)
{
    const char* names[] = { "AnimationTime", "Enabled" };

    // getPropertyHandles expects equal sizes
    GTF_OSAL_ASSERT(ARRAY_SIZE(names) == ARRAY_SIZE(propertyHandles));

    // For accessing the properties uint16_t handles are used.
    // Resolving the handles by the SCDR name of the properties once at startup.
    registration_.getPropertyHandles(names, LastHandle, propertyHandles);

    // Allocate memory for the custom renderer by using the SCDR memory management.
    void* mem = context_.alloc(sizeof(DirectX11RippleEffect));
    // Use placement new to construct the renderer.
    DirectX11RippleEffect* d3d11Renderer = new (mem)DirectX11RippleEffect();
    if (!d3d11Renderer)
    {
        return NULL;
    }

    // Initialize the custom renderer
    if (!d3d11Renderer->init(context_))
    {
        // Cleanup on failure
        d3d11Renderer->~DirectX11RippleEffect();
        context_.free(d3d11Renderer);
        d3d11Renderer = NULL;
    }

    return d3d11Renderer;
}

// Callback for destroying a renderer. This callback is set in 
// RippleEffectPlugin::registerRenderer
void DirectX11RippleEffect::destroyInstance(
    gtf::scdr::sdk::WidgetRenderer* renderer_,
    const gtf::scdr::sdk::WidgetRendererRegistration&,
    gtf::scdr::sdk::GraphicsContext& context_,
    void*)
{
    GTF_OSAL_ASSERT(renderer_);
    if (renderer_)
    {
        // Downcast the renderer to a pointer of the custom renderer.
        DirectX11RippleEffect* renderer = static_cast<DirectX11RippleEffect*>(renderer_);

        // De-initialize the renderer
        renderer->deinit(context_);

        // Call the destructor manually because the memory deallocation does not take care of object destruction.
        renderer->~DirectX11RippleEffect();

        // Free the renderer using the SCDR memory management
        context_.free(renderer_);
    }
}

DirectX11RippleEffect::DirectX11RippleEffect()
    : renderListHead(NULL)
    , renderListTrail(NULL)

{
    memset(&userData, 0, sizeof(UserData));
}

DirectX11RippleEffect::~DirectX11RippleEffect()
{
}

bool DirectX11RippleEffect::init(gtf::scdr::sdk::GraphicsContext& context_)
{
    // The init method is called once at the instantiation of the renderer. 
    // Use the init method to initialize all graphics resources which do not
    // change during the lifetime of the renderer.
    // Calls to the graphics API must not be issued in the methods inherited by
    // gtf::scdr::sdk::WidgetRenderer. 
    // executeUpdateFunction shall be used to initialize graphics resources.
    // The parameter doInit is a pointer to a function implementing the custom specific
    // OpenGL code for initialization. User data is a pointer passed to the doInit
    // function.
    context_.executeUpdateFunction(doInit, &userData);
    return true;
}

void DirectX11RippleEffect::deinit(gtf::scdr::sdk::GraphicsContext& context_)
{
    // The deinit method is called once at the destruction of the renderer.
    // Use the deinit method to destroy graphics resources at the end of the lifetime 
    // of the renderer.
    // Calls to the graphics API must not be issued in the methods inherited by
    // gtf::scdr::sdk::WidgetRenderer.
    // executeDeleteFunction shall be used to destroy graphics resources.
    // The parameter doDeinit is a pointer to a function implementing the custom specific
    // OpenGL code for de-initialization. User data is a pointer passed to the doDeinit
    // function.
    context_.executeDeleteFunction(doDeinit, &userData);
}

bool DirectX11RippleEffect::create(void* renderData_, gtf::scdr::sdk::GraphicsContext&)
{
    // The create method is called once for every widget instance created in the system. 
    // Use this method to do initialization tasks per widget.
    // The pointer renderData points to the user specific data of this widget instance.
    // The memory referenced by render data is provided by the SCDR. The size of the memory block 
    // is configured in RippleEffectPlugin::registerRenderer. It has the size of 
    // DirectX11RippleEffect::RenderDataSize, which is the size of RenderData.
    // Do not free this memory, it is managed by the SCDR.
    RenderData* renderData = static_cast<RenderData*>(renderData_);
    memset(renderData, 0, sizeof(RenderData));
    return true;
}

bool DirectX11RippleEffect::destroy(void* renderData_, gtf::scdr::sdk::GraphicsContext& context_)
{
    // The destroy method is called once when an instance of a widget is destroyed. 
    // Use this method to do de-initialization tasks per widget.
    RenderData* renderData = static_cast<RenderData*>(renderData_);

    if (NULL != renderData->offscreenFramebuffer)
    {
        context_.destroyFramebuffer(renderData->offscreenFramebuffer);
    }
    renderData->offscreenFramebuffer = NULL;
    renderData->offscreenFramebufferWidth = 0;
    renderData->offscreenFramebufferHeight = 0;

    memset(renderData, 0, sizeof(RenderData));
    return true;
}

bool DirectX11RippleEffect::begin(gtf::scdr::sdk::GraphicsContext&)
{
    // This method is called at the beginning of each frame rendered by the SCDR.
    // Reset the internal chain of custom widget data blocks. Since the memory blocks 
    // behind these pointers are managed by the SCDR there is no need for cleanup here.
    renderListHead = NULL;
    renderListTrail = NULL;
    return true;
}

int32_t DirectX11RippleEffect::traverse(void* renderData_, gtf::scdr::sdk::Widget& widget_, gtf::scdr::sdk::TraverserContext& traverserContext_, gtf::scdr::sdk::GraphicsContext& graphicsContext_)
{
    // When the widget is invisible, so are the children, hence abort traversal
    TRAVERSE_ABORT_AND_DISCARD_ON_ERROR(gtf::scdr::sdk::IsWidgetVisible(widget_));

    // cast the user memory to the RenderData structure
    RenderData* renderData = static_cast<RenderData*>(renderData_);

    // The widget_ interface provides access to the widget data. Widget data are the widget 
    // properties filled in EB GUIDE Studio and additional data like the current framebuffer, 
    // transformation matrices.
    const gtf::scdr::sdk::Widget* parentWidget = widget_.getParent();
    gtf::math::float4x4 transform = gtf::math::float4x4::Identity();
    gtf::math::float4x4 projMat = gtf::math::float4x4::Identity();

    GTF_OSAL_ASSERT(parentWidget);

    // Get the transformation matrices, viewport and framebuffer from the parent widget.
    gtf::scdr::sdk::GetWidgetTransformation(*parentWidget, transform);
    TRAVERSE_ABORT_AND_DISCARD_ON_ERROR(
        gtf::scdr::sdk::GetWidgetViewProjectionMatrix(*parentWidget, projMat));
    TRAVERSE_ABORT_AND_DISCARD_ON_ERROR(
        gtf::scdr::sdk::GetWidgetViewport(*parentWidget, renderData->viewport));
    TRAVERSE_ABORT_AND_DISCARD_ON_ERROR(
        gtf::scdr::sdk::GetWidgetFrameBuffer(
        *parentWidget, reinterpret_cast<void**>(&renderData->framebuffer),
        renderData->framebufferId));

    // Get the widget position and size of the current widget.
    // The position is relative to the position of the parent widget.
    int32_t x = 0, y = 0, width = 0, height = 0;
    TRAVERSE_ABORT_AND_DISCARD_ON_ERROR(
        gtf::scdr::sdk::GetWidgetPositionAndExtents(
        widget_, x, y, width, height));

    // Get the z-order of the widget, defining the drawing order of this widget
    const uint32_t depth = widget_.getZOrder();

    int32_t offsetX = 0, offsetY = 0;

    // If this call fails, offsetX and offsetY are not modified
    gtf::scdr::sdk::GetWidgetViewportOffsets(*widget_.getParent(), offsetX, offsetY);

    const float fposx = static_cast<float>(x)+static_cast<float>(offsetX);
    const float fposy = static_cast<float>(y)+static_cast<float>(offsetY);
    const float fwidth = static_cast<float>(width);
    const float fheight = static_cast<float>(height);
    const float fdepth = static_cast<float>(depth);

    // The transformation matrix of this widget.
    const gtf::math::float4x4 widgetT = transform *
        gtf::math::float4x4::Translation(fposx, fposy, fdepth) *
        gtf::math::float4x4::Scaling2D(fwidth, fheight);

    // Set the matrices in the constant buffer data.
    SetConstantBufferMatrix(widgetT, renderData->cbViewPojectionData.viewMat);
    SetConstantBufferMatrix(projMat, renderData->cbViewPojectionData.projMat);

    // Create or update a framebuffer
    TRAVERSE_ABORT_AND_DISCARD_ON_ERROR(
        createOrUpdateFramebuffer(static_cast<int16_t>(width), static_cast<int16_t>(height), *renderData, graphicsContext_));
    renderData->offscreenFramebufferId = traverserContext_.createFramebufferId();

    // Calculate the projection matrix for the widgets rendered to the framebuffer
    const gtf::math::float4x4 childProjMat =
        calcFramebufferProjectionMatrix(static_cast<int16_t>(width), static_cast<int16_t>(height), fdepth);

    // The transformation matrices, the viewport and the framebuffer must be set in the 
    // current widget to make it available for child widgets. 
    // Otherwise the traversal step of any child widget will fail, models will not work properly.
    TRAVERSE_ABORT_AND_DISCARD_ON_ERROR(gtf::scdr::sdk::SetWidgetViewProjectionMatrix(widget_, childProjMat));

    // The transformation matrix for the widgets rendered to the framebuffer.
    // Rendering to the framebuffer starts with an identity matrix.
    const gtf::math::float4x4 rippleTransform = gtf::math::float4x4::Identity();
    gtf::scdr::sdk::SetWidgetTransformation(widget_, rippleTransform);

    const int16_t offscreenViewport[] = { 0, 0, static_cast<int16_t>(width), static_cast<int16_t>(height) };
    TRAVERSE_ABORT_AND_DISCARD_ON_ERROR(gtf::scdr::sdk::SetWidgetViewport(widget_, offscreenViewport));

    // Set the off screen frame buffer for the current widget. This will be read and used by the child widgets.
    // Using the off screen framebuffer id ensures that the content of the framebuffer will be rendered before
    // this widget.
    TRAVERSE_ABORT_AND_DISCARD_ON_ERROR(
        gtf::scdr::sdk::SetWidgetFrameBuffer(widget_, renderData->offscreenFramebuffer, renderData->offscreenFramebufferId));

    renderData->drawId = static_cast<uint16_t>(depth);

    // Custom properties
    bool rippleEnabled = false;
    int32_t rippleTime = 0;

    // Ripple time
    TRAVERSE_ABORT_AND_DISCARD_ON_ERROR(
        widget_.getValue(propertyHandles[Time], GtfTypeId::eTypeId_int32, 1, &(rippleTime), sizeof(uint32_t)));

    // Ripple enabled
    TRAVERSE_ABORT_AND_DISCARD_ON_ERROR(
        widget_.getValue(propertyHandles[Enabled], GtfTypeId::eTypeId_bool, 1, &(rippleEnabled), sizeof(bool)));

    renderData->cbRippleData.active = (rippleEnabled) ? 1.0f : 0.0f;
    renderData->cbRippleData.time = static_cast<float>(rippleTime);

    // Append the render data of this widget to the linked list.
    renderListAppend(renderData);

    return TRAVERSE_SUCCESS;
}

bool DirectX11RippleEffect::update(gtf::scdr::sdk::GraphicsContext&)
{
    // Here is typically code issued which updates graphics resources like
    // VBOs or textures. Graphics code for updating this resources would be issued 
    // with a GraphicsContext::executeUpdateFunction command.
    // Nothing to do for the ripple effect widget here.
    return true;
}

bool DirectX11RippleEffect::render(gtf::scdr::sdk::GraphicsContext& context_)
{
    // Walk over the linked list with the render data and issue the rendering for each object
    for (RenderData* renderData = renderListHead; renderData; renderData = renderData->next)
    {
        // Use the SCDR allocFrame to create a RenderFunctionContext structure. 
        RenderFunctionContext* renderFunctionContext = static_cast<RenderFunctionContext*>(context_.allocFrame(sizeof(RenderFunctionContext)));

        if (NULL == renderFunctionContext)
        {
            return false;
        }

        // For rendering the per widget render data (RenderData) and the common data 
        // for all instances (UserData) are needed.
        renderFunctionContext->renderData = renderData;
        renderFunctionContext->userData = &userData;
        const bool isOpaque = false;

        // Generate a scheduler key for the clear framebuffer command and issue doClear()
        gtf::scdr::sdk::ScheduleDescriptor scheduleClear = gtf::scdr::sdk::CreateClearScheduleDescriptor(renderData->offscreenFramebufferId);
        context_.executeRenderFunction(doClear, renderData, scheduleClear);

        // Generate a scheduler key for rendering. The SCDR ensures that all internal and external widgets
        // are rendered in a correct and fast order. The data in the schedule key is used for
        // sorting the render jobs.
        gtf::scdr::sdk::CScheduleDescriptor scheduleRender(renderData->framebufferId, renderData->drawId, isOpaque);

        // Issue a the delete function doDelete for the RenderFunctionContext structure.
        // Use the executeDeleteFunction to ensure that the renderFunctionContext remains 
        // alive during rendering executed by executeRenderFunction.
        if (false == context_.executeDeleteFunction(doDelete, renderFunctionContext))
        {
            // In case the execute delete function was not issued successfully
            // free the memory in place and return an error.
            context_.freeFrame(renderFunctionContext);
            return false;
        }

        // Execute rendering by issuing the executeRenderFunction. 
        // Parameters are 
        // * the doRender function, which implements the rendering
        // * the renderFunctionContext structure with the render data
        // * the schedule structure with the schedule data.
        context_.executeRenderFunction(doRender, renderFunctionContext, scheduleRender);
    }

    return true;
}

bool DirectX11RippleEffect::end(gtf::scdr::sdk::GraphicsContext&)
{
    // This method is called at the end of each frame rendered.
    // Reset the internal chain of custom widget data blocks. Since the memory blocks 
    // behind these pointers are managed by the SCDR there is no need for cleanup here.
    renderListHead = NULL;
    renderListTrail = NULL;
    return true;
}

bool DirectX11RippleEffect::doInit(void* userData_, gtf::scdr::sdk::GraphicsExecutionContext& context_)
{
    // Implementation of the initialization code of the renderer.
    // This method is issued by DirectX11RippleEffect::init. Since 
    // userData_ is a pointer to UserData, set in DirectX11RippleEffect::init,
    // the void pointer can be casted here.
    if (NULL == userData_)
    {
        return false;
    }

    UserData* data = static_cast<UserData*>(userData_);
    void* dev = NULL;
    context_.getHandle(gtf::scdr::sdk::GraphicsHandles::D3D11Device, dev);
    ID3D11Device* d3dDevice = static_cast<ID3D11Device*>(dev);

    bool success = createShader(data, d3dDevice);
    success = (success) ? createVbo(data, d3dDevice) : false;
    success = (success) ? createCbo(data, d3dDevice) : false;
    success = (success) ? createDepthStencilState(data, d3dDevice) : false;
    success = (success) ? createRasterizerState(data, d3dDevice) : false;
    success = (success) ? createBlendState(data, d3dDevice) : false;
    success = (success) ? createSamplerState(data, d3dDevice) : false;

    if (false == success)
    {
        doDeinit(userData_, context_);
    }

    return success;
}

bool DirectX11RippleEffect::doRender(void* userData_, gtf::scdr::sdk::GraphicsExecutionContext& context_)
{
    // Implementation of the render code. This code was issued by DirectX11RippleEffect::render
    // once for each widget to render.
    if (NULL == userData_)
    {
        return false;
    }

    // Cast the user data to the user specific data structures.
    RenderFunctionContext* renderFunctionContext = static_cast<RenderFunctionContext*>(userData_);
    UserData* userData = renderFunctionContext->userData;
    RenderData* renderData = renderFunctionContext->renderData;

    if (NULL == userData || NULL == renderData || NULL == renderData->framebuffer)
    {
        return false;
    }

    // Set the system framebuffer
    context_.setFramebuffer(renderData->framebuffer);

    void* dev = NULL;
    // Get the device context from the SCDR renderer
    context_.getHandle(gtf::scdr::sdk::GraphicsHandles::D3D11DeviveContext, dev);
    ID3D11DeviceContext* d3dDeviceContext = static_cast<ID3D11DeviceContext*>(dev);

    bool success = checkUserData(userData);
    success = (success) ? updateAndSetCbo(userData, renderData, d3dDeviceContext) : false;

    if (success)
    {
        ID3D11RenderTargetView* renderTargetView = NULL;
        ID3D11DepthStencilView* depthStencilView = NULL;

        ID3D11ShaderResourceView* srcShaderResourceView = NULL;

        // Get handles of render target view, depth stencil view and shader resource view from the SCDR renderer
        context_.getHandle(gtf::scdr::sdk::GraphicsHandles::D3D11RenderTargetView, renderData->framebuffer, reinterpret_cast<void*&>(renderTargetView));
        context_.getHandle(gtf::scdr::sdk::GraphicsHandles::D3D11DepthStencilView, renderData->framebuffer, reinterpret_cast<void*&>(depthStencilView));
        context_.getHandle(gtf::scdr::sdk::GraphicsHandles::D3D11ShaderResourceView, renderData->offscreenFramebuffer, reinterpret_cast<void*&>(srcShaderResourceView));

        d3dDeviceContext->OMSetRenderTargets(1, &renderTargetView, depthStencilView);

        UINT stride = sizeof(float) * 4;
        UINT vboffset = 0;

        d3dDeviceContext->IASetVertexBuffers(0, 1, &(userData->vertexBuffer), &stride, &vboffset);
        d3dDeviceContext->IASetInputLayout(userData->inputLayout);
        d3dDeviceContext->IASetPrimitiveTopology(D3D11_PRIMITIVE_TOPOLOGY_TRIANGLESTRIP);
        d3dDeviceContext->VSSetShader(userData->vertexShader, 0, 0);
        d3dDeviceContext->PSSetShader(userData->pixelShader, 0, 0);

        // Set the shader resource view as texture for the current widget. 
        d3dDeviceContext->PSSetShaderResources(0, 1, &srcShaderResourceView);
        d3dDeviceContext->PSSetSamplers(0, 1, &(userData->samplerState));

        updateViewport(renderData, context_, d3dDeviceContext);
        const bool hasAlpha = true;
        updatePipelineState(userData, d3dDeviceContext, hasAlpha);

        d3dDeviceContext->Draw(4, 0);
    }

    return success;
}

bool DirectX11RippleEffect::doClear(void* renderData_, gtf::scdr::sdk::GraphicsExecutionContext& context_)
{
    if (NULL == renderData_)
    {
        return false;
    }

    // Implementation of the clear framebuffer.
    RenderData* renderData = static_cast<RenderData*>(renderData_);

    ID3D11DeviceContext* dev = NULL;
    ID3D11RenderTargetView* rtv = NULL;
    ID3D11DepthStencilView* dsv = NULL;

    // Get the required DirectX11 handles from the SCDR
    context_.getHandle(gtf::scdr::sdk::GraphicsHandles::D3D11DeviveContext, reinterpret_cast<void*&>(dev));
    context_.getHandle(gtf::scdr::sdk::GraphicsHandles::D3D11RenderTargetView, renderData->offscreenFramebuffer, reinterpret_cast<void*&>(rtv));
    context_.getHandle(gtf::scdr::sdk::GraphicsHandles::D3D11DepthStencilView, renderData->offscreenFramebuffer, reinterpret_cast<void*&>(dsv));

    if (NULL == renderData->offscreenFramebuffer ||
        NULL == dev || NULL == rtv || NULL == dsv)
    {
        return false;
    }

    context_.setFramebuffer(renderData->offscreenFramebuffer);

    const uint32_t clearFlags = D3D11_CLEAR_DEPTH;
    const FLOAT clearColor[4] = { 0.0, 0.0, 0.0, 0.0 };
    const FLOAT clearDepth = 1.0f;
    const UINT8 clearStencil = 0;

    dev->ClearRenderTargetView(rtv, clearColor);
    dev->ClearDepthStencilView(dsv, clearFlags, clearDepth, clearStencil);

    return true;
}

bool DirectX11RippleEffect::doDeinit(void* userData_, gtf::scdr::sdk::GraphicsExecutionContext&)
{
    if (NULL == userData_)
    {
        return false;
    }

    // Implementation of the deinit code. This code is issued by DirectX11RippleEffect::deinit
    UserData* userData = static_cast<UserData*>(userData_);

    if (userData->vertexShader) userData->vertexShader->Release();
    if (userData->pixelShader) userData->pixelShader->Release();
    if (userData->inputLayout) userData->inputLayout->Release();
    if (userData->vertexBuffer) userData->vertexBuffer->Release();
    if (userData->cbViewPojection) userData->cbViewPojection->Release();
    if (userData->cbRipple) userData->cbRipple->Release();
    if (userData->depthStencisState) userData->depthStencisState->Release();
    if (userData->rasterizerState) userData->rasterizerState->Release();
    if (userData->opaqueBlendState) userData->opaqueBlendState->Release();
    if (userData->transparentBlendState) userData->transparentBlendState->Release();
    if (userData->samplerState) userData->samplerState->Release();

    memset(userData, 0, sizeof(UserData));
    return true;
}

bool DirectX11RippleEffect::doDelete(void* memory_, gtf::scdr::sdk::GraphicsExecutionContext& context_)
{
    // Implementation of the doDelete. This code is issued by DirectX11RippleEffect::render to free the 
    // RenderFunctionContext structure
    context_.freeFrame(memory_);
    return true;
}

bool DirectX11RippleEffect::createDepthStencilState(UserData* userData_, ID3D11Device* d3dDevice_)
{
    D3D11_DEPTH_STENCIL_DESC depthStencilDesc;
    ZeroMemory(&depthStencilDesc, sizeof(depthStencilDesc));
    depthStencilDesc.DepthEnable = TRUE;
    depthStencilDesc.DepthWriteMask = D3D11_DEPTH_WRITE_MASK_ALL;
    depthStencilDesc.DepthFunc = D3D11_COMPARISON_LESS;
    depthStencilDesc.StencilEnable = FALSE;
    depthStencilDesc.StencilWriteMask = 0xff;
    depthStencilDesc.StencilReadMask = 0xff;
    depthStencilDesc.FrontFace.StencilFailOp = D3D11_STENCIL_OP_KEEP;
    depthStencilDesc.FrontFace.StencilDepthFailOp = D3D11_STENCIL_OP_KEEP;
    depthStencilDesc.FrontFace.StencilPassOp = D3D11_STENCIL_OP_KEEP;
    depthStencilDesc.FrontFace.StencilFunc = D3D11_COMPARISON_ALWAYS;
    depthStencilDesc.BackFace.StencilFailOp = D3D11_STENCIL_OP_KEEP;
    depthStencilDesc.BackFace.StencilDepthFailOp = D3D11_STENCIL_OP_KEEP;
    depthStencilDesc.BackFace.StencilPassOp = D3D11_STENCIL_OP_KEEP;
    depthStencilDesc.BackFace.StencilFunc = D3D11_COMPARISON_ALWAYS;

    const HRESULT hr = d3dDevice_->CreateDepthStencilState(&depthStencilDesc, &(userData_->depthStencisState));

    return SUCCEEDED(hr);
}

bool DirectX11RippleEffect::createRasterizerState(UserData* userData_, ID3D11Device* d3dDevice_)
{
    D3D11_RASTERIZER_DESC rasterizerDesc;
    ZeroMemory(&rasterizerDesc, sizeof(rasterizerDesc));
    rasterizerDesc.FillMode = D3D11_FILL_SOLID;
    rasterizerDesc.CullMode = D3D11_CULL_NONE;
    rasterizerDesc.FrontCounterClockwise = TRUE;
    rasterizerDesc.DepthBias = 0;
    rasterizerDesc.DepthBiasClamp = 0.0f;
    rasterizerDesc.SlopeScaledDepthBias = 0.0f;
    rasterizerDesc.DepthClipEnable = TRUE;
    rasterizerDesc.ScissorEnable = FALSE;
    rasterizerDesc.MultisampleEnable = FALSE;
    rasterizerDesc.AntialiasedLineEnable = FALSE;

    HRESULT hr = d3dDevice_->CreateRasterizerState(&rasterizerDesc, &(userData_->rasterizerState));

    return SUCCEEDED(hr);
}

bool DirectX11RippleEffect::createBlendState(UserData* userData_, ID3D11Device* d3dDevice_)
{
    D3D11_RENDER_TARGET_BLEND_DESC renderTargetBlendDesc;
    ZeroMemory(&renderTargetBlendDesc, sizeof(renderTargetBlendDesc));
    renderTargetBlendDesc.BlendEnable = TRUE;
    renderTargetBlendDesc.BlendOp = D3D11_BLEND_OP_ADD;
    renderTargetBlendDesc.BlendOpAlpha = D3D11_BLEND_OP_ADD;
    renderTargetBlendDesc.SrcBlend = D3D11_BLEND_SRC_ALPHA;
    renderTargetBlendDesc.SrcBlendAlpha = D3D11_BLEND_ONE;
    renderTargetBlendDesc.DestBlend = D3D11_BLEND_INV_SRC_ALPHA;
    renderTargetBlendDesc.DestBlendAlpha = D3D11_BLEND_INV_SRC_ALPHA;
    renderTargetBlendDesc.RenderTargetWriteMask = 15;

    D3D11_BLEND_DESC blendDesc;
    ZeroMemory(&blendDesc, sizeof(blendDesc));
    blendDesc.AlphaToCoverageEnable = FALSE;
    blendDesc.IndependentBlendEnable = FALSE;

    for (int32_t i = 0; i < ARRAY_SIZE(blendDesc.RenderTarget); i++)
    {
        blendDesc.RenderTarget[i] = renderTargetBlendDesc;
    }

    HRESULT hr = d3dDevice_->CreateBlendState(&blendDesc, &(userData_->transparentBlendState));

    if (FAILED(hr))
    {
        return false;
    }

    renderTargetBlendDesc.BlendEnable = FALSE;
    renderTargetBlendDesc.BlendOp = D3D11_BLEND_OP_ADD;
    renderTargetBlendDesc.BlendOpAlpha = D3D11_BLEND_OP_ADD;
    renderTargetBlendDesc.SrcBlend = D3D11_BLEND_ONE;
    renderTargetBlendDesc.SrcBlendAlpha = D3D11_BLEND_ZERO;
    renderTargetBlendDesc.DestBlend = D3D11_BLEND_ZERO;
    renderTargetBlendDesc.DestBlendAlpha = D3D11_BLEND_ONE;

    for (int32_t i = 0; i < ARRAY_SIZE(blendDesc.RenderTarget); i++)
    {
        blendDesc.RenderTarget[i] = renderTargetBlendDesc;
    }

    hr = d3dDevice_->CreateBlendState(&blendDesc, &(userData_->opaqueBlendState));

    return SUCCEEDED(hr);
}

bool DirectX11RippleEffect::createSamplerState(UserData* userData_, ID3D11Device* d3dDevice_)
{
    D3D11_SAMPLER_DESC desc;
    ZeroMemory(&desc, sizeof(desc));
    desc.Filter = D3D11_FILTER_MIN_MAG_MIP_LINEAR;
    desc.AddressU = D3D11_TEXTURE_ADDRESS_CLAMP;
    desc.AddressV = D3D11_TEXTURE_ADDRESS_CLAMP;
    desc.AddressW = D3D11_TEXTURE_ADDRESS_CLAMP;
    desc.MipLODBias = 0.0f;
    desc.MaxAnisotropy = 0;
    desc.ComparisonFunc = D3D11_COMPARISON_NEVER;
    desc.MinLOD = 0;
    desc.MaxLOD = D3D11_FLOAT32_MAX;

    const HRESULT hr = d3dDevice_->CreateSamplerState(&desc, &(userData_->samplerState));

    return SUCCEEDED(hr);
}

bool DirectX11RippleEffect::createCbo(UserData* userData_, ID3D11Device* d3dDevice_)
{
    D3D11_BUFFER_DESC cbBufDesc;
    ZeroMemory(&cbBufDesc, sizeof(cbBufDesc));
    cbBufDesc.ByteWidth = sizeof(ViewProjectionContstantBuffer);
    cbBufDesc.Usage = D3D11_USAGE_DYNAMIC;
    cbBufDesc.BindFlags = D3D11_BIND_CONSTANT_BUFFER;
    cbBufDesc.CPUAccessFlags = D3D11_CPU_ACCESS_WRITE;

    HRESULT hr = d3dDevice_->CreateBuffer(&cbBufDesc, NULL, &(userData_->cbViewPojection));

    if (FAILED(hr))
    {
        return false;
    }

    cbBufDesc.ByteWidth = sizeof(RippleParametersConstantBuffer);
    hr = d3dDevice_->CreateBuffer(&cbBufDesc, NULL, &(userData_->cbRipple));

    return SUCCEEDED(hr);
}

bool DirectX11RippleEffect::createVbo(UserData* userData_, ID3D11Device* d3dDevice_)
{
    static float vertices[] =
    {
        0.f, 0.f, 0.f, 0.f,  // pos.x, pos,y, tex.u, tex.v
        0.f, 1.f, 0.f, 1.f,
        1.f, 0.f, 1.f, 0.f,
        1.f, 1.f, 1.f, 1.f
    };

    D3D11_BUFFER_DESC bufDesc;
    ZeroMemory(&bufDesc, sizeof(bufDesc));
    bufDesc.ByteWidth = sizeof(float) * _countof(vertices);
    bufDesc.Usage = D3D11_USAGE_IMMUTABLE;
    bufDesc.BindFlags = D3D11_BIND_VERTEX_BUFFER;

    D3D11_SUBRESOURCE_DATA bufData;
    ZeroMemory(&bufData, sizeof(bufData));
    bufData.pSysMem = vertices;

    const HRESULT hr = d3dDevice_->CreateBuffer(&bufDesc, &bufData, &(userData_->vertexBuffer));

    return SUCCEEDED(hr);
}

bool DirectX11RippleEffect::createShader(UserData* userData_, ID3D11Device* d3dDevice_)
{
    const char* msg = NULL;
    ID3DBlob* errorBlob = NULL;
    ID3DBlob* vertexShaderBlob = NULL;
    ID3DBlob* pixelShaderBlob = NULL;
    char vTarget[] = "vs_5_0";
    char pTarget[] = "ps_5_0";
    uint32_t flags = 0;

    // Vertex shader
    HRESULT hr = D3DCompile(VertexShader, strlen(VertexShader), NULL, NULL, NULL, "main", vTarget, flags, 0, &vertexShaderBlob, &errorBlob);
    if (FAILED(hr))
    {
        msg = static_cast<const char*>(errorBlob->GetBufferPointer());
        fprintf(stderr, "%s", msg);
        return false;
    }
    hr = d3dDevice_->CreateVertexShader(vertexShaderBlob->GetBufferPointer(), vertexShaderBlob->GetBufferSize(), 0, &(userData_->vertexShader));
    if (FAILED(hr))
    {
        return false;
    }

    // Pixel shader
    hr = D3DCompile(PixelShader, strlen(PixelShader), NULL, NULL, NULL, "main", pTarget, flags, 0, &(pixelShaderBlob), &errorBlob);
    if (FAILED(hr))
    {
        msg = static_cast<const char*>(errorBlob->GetBufferPointer());
        fprintf(stderr, "%s", msg);
        return false;
    }
    hr = d3dDevice_->CreatePixelShader(pixelShaderBlob->GetBufferPointer(), pixelShaderBlob->GetBufferSize(), 0, &(userData_->pixelShader));
    if (FAILED(hr))
    {
        return false;
    }

    // Input layout
    D3D11_INPUT_ELEMENT_DESC inputDesc[] =
    {
        { "POSITION", 0, DXGI_FORMAT_R32G32_FLOAT, 0, 0, D3D11_INPUT_PER_VERTEX_DATA, 0 },
        { "TEXCOORD", 0, DXGI_FORMAT_R32G32_FLOAT, 0, 2 * sizeof(float), D3D11_INPUT_PER_VERTEX_DATA, 0 }
    };

    uint32_t size = _countof(inputDesc);
    hr = d3dDevice_->CreateInputLayout(inputDesc, size, vertexShaderBlob->GetBufferPointer(), vertexShaderBlob->GetBufferSize(), &(userData_->inputLayout));

    if (NULL != errorBlob)
    {
        errorBlob->Release();
    }

    return true;
}

bool DirectX11RippleEffect::createOrUpdateFramebuffer(uint16_t width_, uint16_t height_, RenderData& renderData_, gtf::scdr::sdk::GraphicsContext& graphicsContext_)
{
    // Avoid framebuffers with size 0/0
    if (0 == width_ || 0 == height_)
    {
        return false;
    }

    // Create a new framebuffer only if width or height has changed
    if ((renderData_.offscreenFramebufferWidth != width_) || (renderData_.offscreenFramebufferHeight != height_))
    {
        if (NULL != renderData_.offscreenFramebuffer)
        {
            graphicsContext_.destroyFramebuffer(renderData_.offscreenFramebuffer);
        }

        gtf::scdr::sdk::FramebufferDescriptor fbDesc;
        fbDesc.width = width_;
        fbDesc.height = height_;
        fbDesc.renderTargetFormat = gtf::scdr::sdk::TextureFormat::R8G8B8A8;
        fbDesc.depthStencilFormat = gtf::scdr::sdk::TextureFormat::D16;
        renderData_.offscreenFramebufferWidth = width_;
        renderData_.offscreenFramebufferHeight = height_;

        return graphicsContext_.createFramebuffer(fbDesc, renderData_.offscreenFramebuffer);
    }

    return true;
}

bool DirectX11RippleEffect::checkUserData(UserData* userData_)
{
    return (NULL != userData_) &&
           (NULL != userData_->vertexShader) &&
           (NULL != userData_->pixelShader) &&
           (NULL != userData_->inputLayout) &&
           (NULL != userData_->vertexBuffer) &&
           (NULL != userData_->cbViewPojection) &&
           (NULL != userData_->cbRipple) &&
           (NULL != userData_->depthStencisState) &&
           (NULL != userData_->rasterizerState) &&
           (NULL != userData_->transparentBlendState) &&
           (NULL != userData_->opaqueBlendState) &&
           (NULL != userData_->samplerState);
}

void DirectX11RippleEffect::updatePipelineState(UserData* userData_, ID3D11DeviceContext* context_, bool hasAlpha_)
{
    FLOAT blendFactor[] = { 0.0f, 0.0f, 0.0f, 0.0f };

    if (hasAlpha_)
    {
        context_->OMSetBlendState(userData_->transparentBlendState, blendFactor, 0xffffff);
    }
    else
    {
        context_->OMSetBlendState(userData_->opaqueBlendState, blendFactor, 0xffffff);
    }
    context_->OMSetDepthStencilState(userData_->depthStencisState, 0);
    context_->RSSetState(userData_->rasterizerState);
}

void DirectX11RippleEffect::updateViewport(RenderData* renderData_, gtf::scdr::sdk::GraphicsExecutionContext& context_, ID3D11DeviceContext* deviceContext_)
{
    D3D11_VIEWPORT vp;
    void* vpPtr = &vp;
    context_.getHandle(gtf::scdr::sdk::GraphicsHandles::Viewport, renderData_->framebuffer, vpPtr);
    deviceContext_->RSSetViewports(1, &vp);
}

bool DirectX11RippleEffect::updateAndSetCbo(UserData* userData_, RenderData* renderData_, ID3D11DeviceContext* context_)
{
    HRESULT hr = S_OK;
    {
        D3D11_MAPPED_SUBRESOURCE mapped;
        const UINT subresource = D3D11CalcSubresource(0, 0, 0);

        // constant buffer for ripple data
        hr = context_->Map(userData_->cbRipple, subresource, D3D11_MAP_WRITE_DISCARD, 0, &mapped);

        if (SUCCEEDED(hr))
        {
            memcpy(mapped.pData, &(renderData_->cbRippleData), sizeof(RippleParametersConstantBuffer));
        }

        context_->Unmap(userData_->cbRipple, subresource);
    }
    {
        D3D11_MAPPED_SUBRESOURCE mapped;
        const UINT subresource = D3D11CalcSubresource(0, 0, 0);

        // constant buffer for view projection matrix
        hr = context_->Map(userData_->cbViewPojection, subresource, D3D11_MAP_WRITE_DISCARD, 0, &mapped);

        if (SUCCEEDED(hr))
        {
            memcpy(mapped.pData, &(renderData_->cbViewPojectionData), sizeof(ViewProjectionContstantBuffer));
        }

        context_->Unmap(userData_->cbViewPojection, subresource);
    }

    if (SUCCEEDED(hr))
    {
        // Update the constant buffers
        context_->VSSetConstantBuffers(0, 1, &(userData_->cbViewPojection));
        context_->PSSetConstantBuffers(0, 1, &(userData_->cbRipple));
    }

    return SUCCEEDED(hr);
}

void DirectX11RippleEffect::renderListAppend(RenderData* data_)
{
    // Append data_ to the internal linked list with the user data.
    (NULL == renderListHead) ? (renderListHead = data_) : (renderListTrail->next = data_);
    renderListTrail = data_;
}

gtf::math::float4x4 DirectX11RippleEffect::calcFramebufferProjectionMatrix(float width_, float height_, float depth_)
{
    const gtf::math::float4x4 flipY = gtf::math::float4x4::Translation(0.f, height_, 0.f) * gtf::math::float4x4::Scaling(1.f, -1.f, -1.f);
    const gtf::math::float4x4 childProjMat = gtf::math::float4x4::D3DOrthographicOffCenterRH(0.f, width_, 0.f, height_, 0.f, depth_);
    const gtf::math::float4x4 projection = childProjMat * flipY;
    
    return projection;
}