////////////////////////////////////////////////////////////////////////////////
// Copyright (c) Elektrobit Automotive GmbH
// Alle Rechte vorbehalten. All Rights Reserved.
//
// Information contained herein is subject to change without notice.
// Elektrobit retains ownership and all other rights in the software and each
// component thereof.
// Any reproduction of the software or components thereof without the prior
// written permission of Elektrobit is prohibited.
////////////////////////////////////////////////////////////////////////////////

#include "GLES20RippleEffect.h"

#include <gtf/scdr/sdk/GraphicsContext.h>
#include <gtf/scdr/sdk/GraphicsExecutionContext.h>
#include <gtf/scdr/sdk/WidgetUtils.h>

#include <GtfTypes/GtfMacros.h>

#include <string.h>
#include <stdio.h>
#include <new>

static const char* VertexShader =
"uniform highp mat4 uProjMat;                                                                               "
"uniform highp mat4 uModelMat;                                                                              "
"                                                                                                           "
"attribute highp vec2 aPosition;                                                                            "
"attribute highp vec2 aTexCoord;                                                                            "
"                                                                                                           "
"varying mediump vec2 vTexCoord;                                                                            "
"                                                                                                           "
"void main()                                                                                                "
"{                                                                                                          "
"    vec4 pos = uModelMat * vec4(aPosition, 0.0, 1.0);                                                      "
"    gl_Position = uProjMat * pos;                                                                          "
"    vTexCoord = aTexCoord;                                                                                 "
"}                                                                                                          ";

static const char* FragmentShader =
"uniform sampler2D uSampler;                                                                                "
"uniform highp float uRippleTime;                                                                           "
"uniform highp float uRippleEnable;                                                                         "
"                                                                                                           "
"varying mediump vec2 vTexCoord;                                                                            "
"                                                                                                           "
"void main()                                                                                                "
"{                                                                                                          "
"    highp vec2 p = -1.0 + 2.0 * vTexCoord;                                                                 "
"    highp float l = length(p);                                                                             "
"    highp vec2 uv = vTexCoord + uRippleEnable * (p / l)*cos(l * 12.0 - 4.0 * uRippleTime / 100.0)*0.03;    "
"    gl_FragColor = texture2D(uSampler, uv);                                                                "
"}                                                                                                          ";

uint16_t GLES20RippleEffect::propertyHandles[GLES20RippleEffect::LastHandle] = { 0, 0 };

// Callback for instantiating a new renderer. This callback is set in 
// RippleEffectPlugin::registerRenderer.
gtf::scdr::sdk::WidgetRenderer* GLES20RippleEffect::createIntance(
    const gtf::scdr::sdk::WidgetRendererRegistration& registration_,
    gtf::scdr::sdk::GraphicsContext& context_,
    void*)
{
    const char* names[] = { "AnimationTime", "Enabled" };

    // getPropertyHandles expects equal sizes
    GTF_OSAL_ASSERT(ARRAY_SIZE(names) == ARRAY_SIZE(propertyHandles));

    // For accessing the properties uint16_t handles are used.
    // Resolving the handles by the SCDR name of the properties once at startup.
    registration_.getPropertyHandles(names, LastHandle, propertyHandles);

    // Allocate memory for the custom renderer by using the SCDR memory management.
    void* mem = context_.alloc(sizeof(GLES20RippleEffect));
    // Use placement new to construct the renderer.
    GLES20RippleEffect* glesRenderer = new (mem)GLES20RippleEffect();
    if (NULL == glesRenderer)
    {
        return NULL;
    }

    // Initialize the custom renderer
    if (!glesRenderer->init(context_))
    {
        // Cleanup on failure
        GLES20RippleEffect::destroyInstance(glesRenderer, registration_, context_, NULL);
        glesRenderer = NULL;
    }

    return glesRenderer;
}

// Callback for destroying a renderer. This callback is set in 
// RippleEffectPlugin::registerRenderer.
void GLES20RippleEffect::destroyInstance(
    gtf::scdr::sdk::WidgetRenderer* renderer_,
    const gtf::scdr::sdk::WidgetRendererRegistration&,
    gtf::scdr::sdk::GraphicsContext& context_,
    void*)
{
    GTF_OSAL_ASSERT(renderer_);
    if (renderer_)
    {
        // Downcast the renderer to a pointer of the custom renderer.
        GLES20RippleEffect* renderer = static_cast<GLES20RippleEffect*>(renderer_);

        // De-initialize the renderer
        renderer->deinit(context_);

        // Call the destructor manually because the memory deallocation does not take care of object destruction.
        renderer->~GLES20RippleEffect();

        // Free the renderer using the SCDR memory management.
        context_.free(renderer_);
    }
}

GLES20RippleEffect::GLES20RippleEffect()
    :gtf::scdr::sdk::WidgetRenderer()
    , renderListHead(NULL)
    , renderListTrail(NULL)
{
    memset(&userData, 0, sizeof(UserData));
}

GLES20RippleEffect::~GLES20RippleEffect()
{
}

bool GLES20RippleEffect::init(gtf::scdr::sdk::GraphicsContext& context_)
{
    // The init method is called once at the instantiation of the renderer. 
    // Use the init method to initialize all graphics resources which do not
    // change during the lifetime of the renderer.
    // Calls to the graphics API must not be issued in the methods inherited by
    // gtf::scdr::sdk::WidgetRenderer. 
    // executeUpdateFunction shall be used to initialize graphics resources.
    // The parameter doInit is a pointer to a function implementing the custom specific
    // OpenGL code for initialization. User data is a pointer passed to the doInit
    // function.
    return context_.executeUpdateFunction(doInit, &userData);
}

void GLES20RippleEffect::deinit(gtf::scdr::sdk::GraphicsContext& context_)
{
    // The deinit method is called once at the destruction of the renderer.
    // Use the deinit method to destroy graphics resources at the end of the lifetime 
    // of the renderer.
    // Calls to the graphics API must not be issued in the methods inherited by
    // gtf::scdr::sdk::WidgetRenderer.
    // executeDeleteFunction shall be used to destroy graphics resources.
    // The parameter doDeinit is a pointer to a function implementing the custom specific
    // OpenGL code for de-initialization. User data is a pointer passed to the doDeinit
    // function.
    context_.executeDeleteFunction(doDeinit, &userData);
}

bool GLES20RippleEffect::create(void* renderData_, gtf::scdr::sdk::GraphicsContext&)
{
    // The create method is called once for every widget instance created in the system. 
    // Use this method to do initialization tasks per widget.
    // The pointer renderData points to the user specific data of this widget instance.
    // The memory referenced by render data is provided by the SCDR. The size of the memory block 
    // is configured in RippleEffectPlugin::registerRenderer. It has the size of 
    // GLES20RippleEffect::RenderDataSize, which is the size of RenderData.
    // Do not free this memory, it is managed by the SCDR.
    RenderData* renderData = static_cast<RenderData*>(renderData_);
    memset(renderData, 0, sizeof(RenderData));
    return true;
}

bool GLES20RippleEffect::destroy(void* renderData_, gtf::scdr::sdk::GraphicsContext& context_)
{
    // The destroy method is called once when an instance of a widget is destroyed. 
    // Use this method to do de-initialization tasks per widget.
    RenderData* renderData = static_cast<RenderData*>(renderData_);
    context_.destroyFramebuffer(renderData->offscreenFramebuffer);
    renderData->offscreenFramebuffer = NULL;
    renderData->offscreenFramebufferWidth = 0;
    renderData->offscreenFramebufferHeight = 0;

    memset(renderData, 0, sizeof(RenderData));
    return true;
}

bool GLES20RippleEffect::begin(gtf::scdr::sdk::GraphicsContext&)
{
    // This method is called at the beginning of each frame rendered by the SCDR.
    // Reset the internal chain of custom widget data blocks. Since the memory blocks 
    // behind these pointers are managed by the SCDR there is no need for cleanup here.
    renderListHead = NULL;
    renderListTrail = NULL;
    return true;
}

int32_t GLES20RippleEffect::traverse(void* renderData_, gtf::scdr::sdk::Widget& widget_, gtf::scdr::sdk::TraverserContext& traverserContext_, gtf::scdr::sdk::GraphicsContext& graphicsContext_)
{
    // When the widget is invisible, so are the children, hence abort traversal.
    TRAVERSE_ABORT_AND_DISCARD_ON_ERROR(gtf::scdr::sdk::IsWidgetVisible(widget_));

    // Cast the user memory to the RenderData structure.
    RenderData* renderData = static_cast<RenderData*>(renderData_);

    // The widget_ interface provides access to the widget data. Widget data are the widget 
    // properties filled in EB GUIDE Studio and additional data like the current framebuffer, 
    // transformation matrices.
    const gtf::scdr::sdk::Widget* parentWidget = widget_.getParent();
    gtf::math::float4x4 transform = gtf::math::float4x4::Identity();
    renderData->projMat = gtf::math::float4x4::Identity();
    void* framebuffer = 0;

    GTF_OSAL_ASSERT(parentWidget);

    // Get the transformation matrices, viewport and framebuffer from the parent widget.
    gtf::scdr::sdk::GetWidgetTransformation(*parentWidget, transform);
    TRAVERSE_ABORT_AND_DISCARD_ON_ERROR(gtf::scdr::sdk::GetWidgetViewProjectionMatrix(*parentWidget, renderData->projMat));
    TRAVERSE_ABORT_AND_DISCARD_ON_ERROR(gtf::scdr::sdk::GetWidgetViewport(*parentWidget, renderData->viewport));
    TRAVERSE_ABORT_AND_DISCARD_ON_ERROR(gtf::scdr::sdk::GetWidgetFrameBuffer(*parentWidget, &framebuffer, renderData->framebufferId));
    renderData->framebuffer = reinterpret_cast<gtf::scdr::sdk::Framebuffer*>(framebuffer);

    // Get the widget position and size of the current widget.
    // The position is relative to the position of the parent widget.
    int32_t x = 0, y = 0, width = 0, height = 0;
    TRAVERSE_ABORT_AND_DISCARD_ON_ERROR(
        gtf::scdr::sdk::GetWidgetPositionAndExtents(
        widget_, x, y, width, height));

    // Get the z-order of the widget, defining the drawing order of this widget.
    const uint32_t depth = widget_.getZOrder();

    int32_t offsetX = 0, offsetY = 0;
    // If this call fails, offsetX and offsetY are not modified, an offset is not applied.
    gtf::scdr::sdk::GetWidgetViewportOffsets(*widget_.getParent(), offsetX, offsetY);

    const float fposx = static_cast<float>(x)+static_cast<float>(offsetX);
    const float fposy = static_cast<float>(y)+static_cast<float>(offsetY);
    const float fwidth = static_cast<float>(width);
    const float fheight = static_cast<float>(height);
    const float fdepth = static_cast<float>(depth);

    // The transformation matrix for the widgets rendered to the framebuffer.
    // Rendering to the framebuffer starts with an identity matrix.
    const gtf::math::float4x4 framebufferTransform = gtf::math::float4x4::Identity();
    gtf::scdr::sdk::SetWidgetTransformation(widget_, framebufferTransform);

    // The transformation matrix for the ripple effect widget.
    renderData->modelMat = transform * gtf::math::float4x4::Translation(fposx, -fposy, fdepth) * gtf::math::float4x4::Scaling2D(fwidth, fheight);
    renderData->drawId = static_cast<uint16_t>(depth);

    // Custom properties
    bool rippleEnabled = false;
    int32_t rippleTime = 0;

    // Ripple time
    TRAVERSE_ABORT_AND_DISCARD_ON_ERROR(
        widget_.getValue(propertyHandles[Time], GtfTypeId::eTypeId_int32, 1, &(rippleTime), sizeof(uint32_t)));

    // Ripple enabled
    TRAVERSE_ABORT_AND_DISCARD_ON_ERROR(
        widget_.getValue(propertyHandles[Enabled], GtfTypeId::eTypeId_bool, 1, &(rippleEnabled), sizeof(bool)));
    renderData->rippleEnable = (rippleEnabled) ? 1.0f : 0.0f;
    renderData->rippleTime = static_cast<float>(rippleTime);

    // Create or update the framebuffer
    TRAVERSE_ABORT_AND_DISCARD_ON_ERROR(
    createOrUpdateFramebuffer(static_cast<int16_t>(width), static_cast<int16_t>(height), *renderData, graphicsContext_));
    // Get an Id for scheduling the renderer calls for the new framebuffer. This id is valid for one frame.
    renderData->offscreenFramebufferId = traverserContext_.createFramebufferId();

    // Calculate the projection matrix for the widgets rendered to the framebuffer
    const gtf::math::float4x4 childProjMat = 
        calcFramebufferProjectionMatrix(static_cast<int16_t>(width), static_cast<int16_t>(height), fdepth);

    // Add a y-flip to the widget projection matrix because the OpenGL|ES 2.0 framebuffer has its origin at bottom left, GTF 
    // has its origin to top left.
    renderData->projMat *= (gtf::math::float4x4::Scaling2D(1.0f, -1.0f) * gtf::math::float4x4::Translation2D(0.0f, -fheight));

    // The transformation matrices, the viewport and the framebuffer must be set in the 
    // current widget to make it available for child widgets. 
    // Otherwise the traversal step of any child widget will fail, models will not work properly.
    TRAVERSE_ABORT_AND_DISCARD_ON_ERROR(
        gtf::scdr::sdk::SetWidgetViewProjectionMatrix(widget_, childProjMat));

    const int16_t offscreenViewport[] = { 0, 0, static_cast<int16_t>(width), static_cast<int16_t>(height) };
    TRAVERSE_ABORT_AND_DISCARD_ON_ERROR(
        gtf::scdr::sdk::SetWidgetViewport(widget_, offscreenViewport));

    // Set the off screen frame buffer for the current widget. This will be read and used by the child widgets.
    // Using the off screen framebuffer id ensures that the content of the framebuffer will be rendered before
    // this widget.
    TRAVERSE_ABORT_AND_DISCARD_ON_ERROR(
        gtf::scdr::sdk::SetWidgetFrameBuffer(widget_, renderData->offscreenFramebuffer, renderData->offscreenFramebufferId));

    // Append the render data of this widget to the linked list.
    renderListAppend(renderData);

    return TRAVERSE_SUCCESS;
}

bool GLES20RippleEffect::update(gtf::scdr::sdk::GraphicsContext&)
{
    // Here is typically code issued which updates graphics resources like
    // VBOs or textures. Graphics code for updating this resources would be issued 
    // with a GraphicsContext::executeUpdateFunction command.
    // Nothing to do for the ripple effect widget here.
    return true;
}

bool GLES20RippleEffect::render(gtf::scdr::sdk::GraphicsContext& context_)
{
    // Walk over the linked list with the render data and issue the rendering for each object.
    for (RenderData* renderData = renderListHead; renderData; renderData = renderData->next)
    {
        // Use the SCDR allocFrame to create a RenderFunctionContext structure. 
        RenderFunctionContext* renderFunctionContext = static_cast<RenderFunctionContext*>(context_.allocFrame(sizeof(RenderFunctionContext)));

        if (NULL == renderFunctionContext)
        {
            // Failed to allocate memory. Stop the rendering and return an error.
            return false;
        }

        // For rendering the per widget render data (RenderData) and the common data 
        // for all instances (UserData) are needed.
        renderFunctionContext->renderData = renderData;
        renderFunctionContext->userData = &userData;
        const bool isOpaque = false;

        // Generate a scheduler key for the clear framebuffer command and issue doClear().
        gtf::scdr::sdk::ScheduleDescriptor scheduleClear = gtf::scdr::sdk::CreateClearScheduleDescriptor(renderData->offscreenFramebufferId);
        context_.executeRenderFunction(doClearFramebuffer, renderData, scheduleClear);

        // Generate a scheduler key for rendering. The SCDR ensures that all internal and external widgets
        // are rendered in a correct and fast order. The data in the schedule key is used for
        // sorting the render jobs.
        gtf::scdr::sdk::CScheduleDescriptor scheduleRender(renderData->framebufferId, renderData->drawId, isOpaque);

        // Issue a the delete function doDelete for the RenderFunctionContext structure.
        // Use the executeDeleteFunction to ensure that the renderFunctionContext remains 
        // alive during rendering executed by executeRenderFunction.
        if (false == context_.executeDeleteFunction(doDelete, renderFunctionContext))
        {
            // In case the execute delete function was not issued successfully
            // free the memory in place and return an error.
            context_.freeFrame(renderFunctionContext);
            return false;
        }

        // Execute rendering by issuing the executeRenderFunction. 
        // Parameters are:
        // * the doRender function, which implements the rendering
        // * the renderFunctionContext structure with the render data
        // * the schedule structure with the schedule data.
        context_.executeRenderFunction(doRender, renderFunctionContext, scheduleRender);
    }

    return true;
}

bool GLES20RippleEffect::end(gtf::scdr::sdk::GraphicsContext&)
{
    // This method is called at the end of each frame rendered.
    // Reset the internal chain of custom widget data blocks. Since the memory blocks 
    // behind these pointers are managed by the SCDR there is no need for cleanup here.
    renderListHead = NULL;
    renderListTrail = NULL;
    return true;
}

bool GLES20RippleEffect::doInit(void* userData_, gtf::scdr::sdk::GraphicsExecutionContext&)
{
    // Implementation of the initialization code of the renderer.
    // This method is issued by GLES20RippleEffect::init. Since 
    // userData_ is a pointer to UserData, set in GLES20RippleEffect::init,
    // the void pointer can be casted here.
    UserData* data = static_cast<UserData*>(userData_);
    return initShader(data->shader, VertexShader, FragmentShader);
}

bool GLES20RippleEffect::doRender(void* userData_, gtf::scdr::sdk::GraphicsExecutionContext& context_)
{
    // Implementation of the render code. This code was issued by GLES20RippleEffect::render
    // once for each widget to render.
    if (NULL == userData_)
    {
        return false;
    }

    // Cast the user data to the user specific data structures.
    RenderFunctionContext* ctx = static_cast<RenderFunctionContext*>(userData_);
    const UserData* userData = ctx->userData;
    const RenderData* renderData = ctx->renderData;

    if (NULL == userData || NULL == renderData || NULL == renderData->framebuffer)
    {
        return false;
    }

    static float vertices[] =
    {
        0.f, 0.f, 0.f, 0.f,  // pos.x, pos.y, tex.u, tex.v
        0.f, 1.f, 0.f, 1.f,
        1.f, 0.f, 1.f, 0.f,
        1.f, 1.f, 1.f, 1.f
    };

    // OpenGL|ES 2.0 implementation of the rendering

    // Disable some render states. Potentially every state can be changed by the SCDR,
    // in every frame. Before rendering the states shall be set to proper values.
    glDisable(GL_SCISSOR_TEST);
    glDisable(GL_STENCIL_TEST);
    glBindBuffer(GL_ARRAY_BUFFER, 0);
    glDisable(GL_CULL_FACE);
    glFrontFace(GL_CCW);
    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LESS);

    // Set the system framebuffer
    context_.setFramebuffer(renderData->framebuffer);

    glViewport(renderData->viewport[0], renderData->viewport[1], renderData->viewport[2], renderData->viewport[3]);

    if (renderData->isOpaque)
    {
        glDisable(GL_BLEND);
    }
    else
    {
        glEnable(GL_BLEND);
        glBlendEquationSeparate(GL_FUNC_ADD, GL_FUNC_ADD);
        glBlendFuncSeparate(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA, GL_ONE, GL_ONE_MINUS_SRC_ALPHA);
        glBlendColor(0.0f, 0.0f, 0.0f, 0.0f);
    }

    glUseProgram(userData->shader.program);

    GLint vertexLocation = glGetAttribLocation(userData->shader.program, "aPosition");

    glVertexAttribPointer(vertexLocation, 2, GL_FLOAT, GL_FALSE, 4 * sizeof(float), vertices);
    const float* texPtr = vertices + 2;
    glEnableVertexAttribArray(vertexLocation);

    GLint texLocation = glGetAttribLocation(userData->shader.program, "aTexCoord");

    glVertexAttribPointer(texLocation, 2, GL_FLOAT, GL_FALSE, 4 * sizeof(float), texPtr);
    glEnableVertexAttribArray(texLocation);

    glUniformMatrix4fv(userData->shader.projMatLocation, 1, GL_FALSE, renderData->projMat.values);
    glUniformMatrix4fv(userData->shader.modelMatLocation, 1, GL_FALSE, renderData->modelMat.values);
    glUniform1f(userData->shader.rippleTimeLocation, renderData->rippleTime);
    glUniform1f(userData->shader.rippleEnabledLocation, renderData->rippleEnable);

    // Get the color buffer of the off screen framebuffer. 
    GLuint texture = 0;
    context_.getHandle(gtf::scdr::sdk::GraphicsHandles::GLES20ColorBufferHandle, renderData->offscreenFramebuffer, reinterpret_cast<void*&>(texture));

    // And use it as texture in the current widget.
    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D, texture);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glUniform1i(userData->shader.framebufferLocation, 0);

    glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);

    const GLenum err = glGetError();
    return (GL_NO_ERROR == err);
}

bool GLES20RippleEffect::doClearFramebuffer(void* renderData_, gtf::scdr::sdk::GraphicsExecutionContext& context_)
{
    // Implementation of the clear framebuffer.
    if (NULL == renderData_)
    {
        return false;
    }

    glDisable(GL_SCISSOR_TEST);

    RenderData* renderData = static_cast<RenderData*>(renderData_);

    if (NULL == renderData->offscreenFramebuffer)
    {
        return false;
    }

    context_.setFramebuffer(renderData->offscreenFramebuffer);

    glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
    glClearDepthf(1.0);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    return true;
}

bool GLES20RippleEffect::doDeinit(void* userData_, gtf::scdr::sdk::GraphicsExecutionContext&)
{
    // Implementation of the deinit code. This code is issued by GLES20RippleEffect::deinit
    if (NULL == userData_)
    {
        return false;
    }

    UserData* data = static_cast<UserData*>(userData_);

    if (data->shader.program && data->shader.fragmentShader) glDetachShader(data->shader.program, data->shader.fragmentShader);
    if (data->shader.program && data->shader.vertexShader) glDetachShader(data->shader.program, data->shader.vertexShader);
    if (data->shader.program) glDeleteProgram(data->shader.program);
    if (data->shader.fragmentShader) glDeleteShader(data->shader.fragmentShader);
    if (data->shader.vertexShader) glDeleteShader(data->shader.vertexShader);

    data->shader.program = 0;
    data->shader.vertexShader = 0;
    data->shader.fragmentShader = 0;

    const GLenum err = glGetError();
    return (GL_NO_ERROR == err);
}

bool GLES20RippleEffect::doDelete(void* memory_, gtf::scdr::sdk::GraphicsExecutionContext& context_)
{
    // Implementation of the doDelete. This code is issued by GLES20RippleEffect::render to free the 
    // RenderFunctionContext structure
    context_.freeFrame(memory_);
    return true;
}

void GLES20RippleEffect::renderListAppend(RenderData* data_)
{
    // Append data_ to the internal linked list with the user data.
    (NULL == renderListHead) ? (renderListHead = data_) : (renderListTrail->next = data_);
    renderListTrail = data_;
}

bool GLES20RippleEffect::initShader(ShaderParams& params_, const char* vertexShader_, const char* fragmentShader_)
{
    // Create and upload a shader program to the GLES20 graphics API
    GLuint vertexShaderHandle = loadShader(GL_VERTEX_SHADER, vertexShader_);
    GLuint fragmentShaderHandle = loadShader(GL_FRAGMENT_SHADER, fragmentShader_);

    GLuint program = glCreateProgram();

    GLint linked = 0;

    if ((0 == vertexShaderHandle) || (0 == fragmentShaderHandle) || (0 == program))
    {
        return false;
    }

    glAttachShader(program, vertexShaderHandle);
    glAttachShader(program, fragmentShaderHandle);

    glBindAttribLocation(program, 0, "aPosition");

    glLinkProgram(program);

    glGetProgramiv(program, GL_LINK_STATUS, &linked);

    if (0 == linked)
    {
        char error[1024];
        glGetProgramInfoLog(program, 1024, NULL, error);
        fprintf(stderr, "%s\n", error);
    }

    params_.projMatLocation = glGetUniformLocation(program, "uProjMat");
    params_.modelMatLocation = glGetUniformLocation(program, "uModelMat");
    params_.rippleTimeLocation = glGetUniformLocation(program, "uRippleTime");
    params_.rippleEnabledLocation = glGetUniformLocation(program, "uRippleEnable");
    params_.framebufferLocation = glGetUniformLocation(program, "uSampler");

    params_.program = program;
    params_.vertexShader = vertexShaderHandle;
    params_.fragmentShader = fragmentShaderHandle;

    return (GL_NO_ERROR == glGetError());
}

GLuint GLES20RippleEffect::loadShader(GLenum type_, const char* shaderSrc_)
{
    // Create and upload a shader to the GLES20 graphics API
    GLint compiled = 0;
    GLuint shader = glCreateShader(type_);

    if (0 == shader)
    {
        return 0;
    }

    glShaderSource(shader, 1, &shaderSrc_, NULL);
    glCompileShader(shader);
    glGetShaderiv(shader, GL_COMPILE_STATUS, &compiled);

    if (0 == compiled)
    {
        char error[1024];
        glGetShaderInfoLog(shader, sizeof(error), NULL, error);
        fprintf(stderr, "%s\n", error);
    }

    return shader;
}

bool GLES20RippleEffect::createOrUpdateFramebuffer(uint16_t width_, uint16_t height_, RenderData& renderData_, gtf::scdr::sdk::GraphicsContext& graphicsContext_)
{
    // Avoid framebuffers with size 0/0
    if (0 == width_ || 0 == height_)
    {
        return false;
    }

    // Create a new framebuffer only if width or height has changed
    if ((renderData_.offscreenFramebufferWidth != width_) || (renderData_.offscreenFramebufferHeight != height_))
    {
        if (NULL != renderData_.offscreenFramebuffer)
        {
            graphicsContext_.destroyFramebuffer(renderData_.offscreenFramebuffer);
        }

        gtf::scdr::sdk::FramebufferDescriptor fbDesc;
        fbDesc.width = static_cast<uint32_t>(width_);
        fbDesc.height = static_cast<uint32_t>(height_);
        fbDesc.renderTargetFormat = gtf::scdr::sdk::TextureFormat::R8G8B8A8;
        fbDesc.depthStencilFormat = gtf::scdr::sdk::TextureFormat::D16;
        renderData_.offscreenFramebufferWidth = width_;
        renderData_.offscreenFramebufferHeight = height_;

        return graphicsContext_.createFramebuffer(fbDesc, renderData_.offscreenFramebuffer);
    }

    return true;
}

gtf::math::float4x4 GLES20RippleEffect::calcFramebufferProjectionMatrix(float width_, float height_, float depth_)
{
    const gtf::math::float4x4 flipZ = gtf::math::float4x4::Scaling(1.0f, 1.0f, -1.0f);
    const gtf::math::float4x4 translate = gtf::math::float4x4::Translation2D(0, static_cast<float>(height_));
    const gtf::math::float4x4 flipX = gtf::math::float4x4::Scaling2D(1.0f, -1.0f);
    const gtf::math::float4x4 projection = gtf::math::float4x4::OGLOrthographicOffCenterRH(0.0f, width_, 0.f, height_, 0.f, depth_);

    return projection * translate * flipX * flipZ;
}