////////////////////////////////////////////////////////////////////////////////
// Copyright (c) Elektrobit Automotive GmbH
// Alle Rechte vorbehalten. All Rights Reserved.
//
// Information contained herein is subject to change without notice.
// Elektrobit retains ownership and all other rights in the software and each
// component thereof.
// Any reproduction of the software or components thereof without the prior
// written permission of Elektrobit is prohibited.
////////////////////////////////////////////////////////////////////////////////

#include "DirectX11TriangleRenderer.h"
#include <gtf/scdr/sdk/GraphicsExecutionContext.h>
#include <gtf/scdr/sdk/GraphicsHandles.h>
#include <gtf/scdr/sdk/Color.h>

#include <d3dcompiler.h>
#include <stdio.h>


static const char* VertexShader =
"cbuffer cbImageParams : register(b0)                     "
"{                                                        "
"    float4x4 projMat;                                    "
"    float4x4 viewMat;                                    "
"    float4   coloration;                                 "
"};                                                       "
"struct VS_IN                                             "
"{                                                        "
"    float2 pos : POSITION0;                              "
"};                                                       "
"struct VS_OUT                                            "
"{                                                        "
"    float4 pos : SV_POSITION;                            "
"    nointerpolation float4 color : COLOR0;               "
"};                                                       "
"VS_OUT main(VS_IN vsIn)                                  "
"{                                                        "
"    VS_OUT vsOut = (VS_OUT)0;                            "
"    float4 wpos = float4(vsIn.pos.x,vsIn.pos.y,0.f,1.f); "
"    float4 viewPos = mul(wpos, viewMat);                 "
"    vsOut.pos = mul(viewPos, projMat);                   "
"    vsOut.color = coloration;                            "
"    return vsOut;                                        "
"};                                                       ";

static const char* PixelShader =
"struct PS_IN                                             "
"{                                                        "
"    float4 pos : SV_POSITION;                            "
"    nointerpolation float4 color : COLOR0;               "
"};                                                       "
"float4 main(PS_IN psIn) : SV_TARGET0                     "
"{                                                        "
"    return psIn.color;                                   "
"}                                                        ";

uint16_t DirectX11TriangleRenderer::colorHandle = 0;

// OpenGL|ES 2.0 and GTF expects column major matrices, DirectX11 expects
// row major matrices. This helper function copies a matrix and converts 
// it for the DirectX11 API.
static void SetConstantBufferMatrix(const gtf::math::float4x4& matrixIn_, FLOAT* cbMatrixOut_)
{
    const float* values = matrixIn_.values;
    for (uint32_t x = 0; x < 4; ++x)
    {
        for (uint32_t y = 0; y < 4; ++y)
        {
            const uint32_t srcIdx = x * 4 + y;
            const uint32_t dstIdx = y * 4 + x;
            cbMatrixOut_[dstIdx] = static_cast<FLOAT>(values[srcIdx]);
        }
    }
}

// Callback for instantiating a new renderer. This callback is set in 
// TrianglePlugin::registerRenderer
gtf::scdr::sdk::WidgetRenderer* DirectX11TriangleRenderer::createIntance(
    const gtf::scdr::sdk::WidgetRendererRegistration& registration_,
    gtf::scdr::sdk::GraphicsContext& context_,
    void*)
{
    const char* names[] = { "Color" };

    // For accessing the properties uint16_t handles are used.
    // Resolving the handles by the SCDR name of the properties once at startup.
    registration_.getPropertyHandles(names, 1, &colorHandle);

    // Allocate memory for the custom renderer by using the SCDR memory management.
    void* mem = context_.alloc(sizeof(DirectX11TriangleRenderer));
    // Use placement new to construct the renderer.
    DirectX11TriangleRenderer* d3d11Renderer = new (mem)DirectX11TriangleRenderer();
    if (!d3d11Renderer)
    {
        return NULL;
    }

    // Initialize the custom renderer
    if (!d3d11Renderer->init(context_))
    {
        // Cleanup on failure
        d3d11Renderer->~DirectX11TriangleRenderer();
        context_.free(d3d11Renderer);
        d3d11Renderer = NULL;
    }

    return d3d11Renderer;
}

// Callback for destroying a renderer. This callback is set in 
// TrianglePlugin::registerRenderer
void DirectX11TriangleRenderer::destroyInstance(
    gtf::scdr::sdk::WidgetRenderer* renderer_,
    const gtf::scdr::sdk::WidgetRendererRegistration&,
    gtf::scdr::sdk::GraphicsContext& context_,
    void*)
{
    GTF_OSAL_ASSERT(renderer_);

    if (renderer_)
    {
        // Downcast the renderer to a pointer of the custom renderer.
        // The type of the renderer is known, the renderer was created in DirectX11TriangleRenderer::createIntance 
        DirectX11TriangleRenderer* renderer = static_cast<DirectX11TriangleRenderer*>(renderer_);

        // De-initialize the renderer
        renderer->deinit(context_);

        // Call the destructor manually because the memory deallocation does not take care of object destruction.
        renderer->~DirectX11TriangleRenderer();

        // Free the renderer using the SCDR memory management
        context_.free(renderer_);
    }
}

DirectX11TriangleRenderer::DirectX11TriangleRenderer()
    : renderListHead(NULL)
    , renderListTrail(NULL)
{
}

DirectX11TriangleRenderer::~DirectX11TriangleRenderer()
{
}

bool DirectX11TriangleRenderer::init(gtf::scdr::sdk::GraphicsContext& context_)
{
    // The init method is called once at the instantiation of the renderer. 
    // Use the init method to initialize all graphics resources which do not
    // change during the lifetime of the renderer.
    // Calls to the graphics API must not be issued in the methods inherited by
    // gtf::scdr::sdk::WidgetRenderer. 
    // executeUpdateFunction shall be used to initialize graphics resources.
    // The parameter doInit is a pointer to a function implementing the custom specific
    // OpenGL code for initialization. User data is a pointer passed to the doInit
    // function.
    return context_.executeUpdateFunction(doInit, &userData);
}

void DirectX11TriangleRenderer::deinit(gtf::scdr::sdk::GraphicsContext& context_)
{
    // The deinit method is called once at the destruction of the renderer.
    // Use the deinit method to destroy graphics resources at the end of the lifetime 
    // of the renderer.
    // Calls to the graphics API must not be issued in the methods inherited by
    // gtf::scdr::sdk::WidgetRenderer.
    // executeDeleteFunction shall be used to destroy graphics resources.
    // The parameter doDeinit is a pointer to a function implementing the custom specific
    // OpenGL code for de-initialization. User data is a pointer passed to the doDeinit
    // function.
    context_.executeDeleteFunction(doDeinit, &userData);
}

bool DirectX11TriangleRenderer::create(void* renderData_, gtf::scdr::sdk::GraphicsContext&)
{
    // The create method is called once for every widget instance created in the system. 
    // Use this method to do initialization tasks per widget.
    // The pointer renderData points to the user specific data of this widget instance.
    // The memory referenced by render data is provided by the SCDR. The size of the memory block 
    // is configured in TrianglePlugin::registerRenderer. It has the size of 
    // GLES20TriangleRenderer::RenderDataSize, which is the size of RenderData.
    // Do not free this memory, it is managed by the SCDR.
    memset(renderData_, 0, sizeof(RenderData));
    return true;
}

bool DirectX11TriangleRenderer::destroy(void* renderData_, gtf::scdr::sdk::GraphicsContext&)
{
    // The destroy method is called once when an instance of a widget is destroyed. 
    // Use this method to do de-initialization tasks per widget.
    memset(renderData_, 0, sizeof(RenderData));
    return true;
}

bool DirectX11TriangleRenderer::begin(gtf::scdr::sdk::GraphicsContext&)
{
    // This method is called at the beginning of each frame rendered by the SCDR.
    // Reset the internal chain of custom widget data blocks. Since the memory blocks 
    // behind these pointers are managed by the SCDR there is no need for cleanup here.
    renderListHead = NULL;
    renderListTrail = NULL;
    return true;
}

int32_t DirectX11TriangleRenderer::traverse(void* renderData_, gtf::scdr::sdk::Widget& widget_, gtf::scdr::sdk::TraverserContext&, gtf::scdr::sdk::GraphicsContext&)
{
    // When the widget is invisible, so are the children, hence abort traversal
    TRAVERSE_ABORT_AND_DISCARD_ON_ERROR(gtf::scdr::sdk::IsWidgetVisible(widget_));

    GTF_OSAL_ASSERT(NULL != renderData_);

    // cast the user memory to the RenderData structure
    RenderData* renderData = static_cast<RenderData*>(renderData_);

    gtf::scdr::sdk::WidgetDefaultProperties properties;

    // Load the default values in the SystemWidgetPropertySets::Full2D
    // Note that the property set Full2D shall be set in TrianglePlugin::registerRenderer
    // to make the default features work correctly.
    // This is an auxiliary function to cover the simple case, no properties handed over to the 
    // child widgets are modified.
    // An example for using the get and set functions for every single property 
    // is shown in the ripple effect example in DirectX11RippleEffect::traverse.
    gtf::scdr::sdk::GetWidgetDefaultProperties(widget_, properties);

    renderData->framebuffer = properties.framebuffer;
    renderData->framebufferId = properties.framebufferId;

    // Set the transformation matrix of this widget, add scaling and z-translation
    const gtf::math::float4x4 modelMat = properties.widgetTransformationMatrix *
        gtf::math::float4x4::Translation(0.0f, 0.0f, static_cast<float>(widget_.getZOrder())) *
        gtf::math::float4x4::Scaling2D(static_cast<float>(properties.width), static_cast<float>(properties.height));

    renderData->drawId = static_cast<uint16_t>(widget_.getZOrder());

    // Copy to constant buffer data structure and convert to
    // row major matrix.
    SetConstantBufferMatrix(modelMat, renderData->cbTriangle.viewMat);
    SetConstantBufferMatrix(properties.viewProjectionMatrix, renderData->cbTriangle.projMat);

    uint32_t uintColor = 0;
    // Get custom property color. In GTF a color is represented as uint32 RGBA value.
    TRAVERSE_ABORT_AND_DISCARD_ON_ERROR(widget_.getValue(colorHandle, GtfTypeId::eTypeId_color, 1, &uintColor, sizeof(uintColor)));
    gtf::scdr::sdk::Color color = gtf::scdr::sdk::UnpackColor(uintColor);

    // Get color from Coloration feature
    gtf::scdr::sdk::Color colorationColor = gtf::scdr::sdk::SystemColors::White;

    // GetAndUpdateWidgetColoration implicitly sets the coloration color for child widgets
    gtf::scdr::sdk::GetAndUpdateWidgetColoration(widget_, colorationColor);

    // Apply the coloration color to the widgets fill color
    color = gtf::scdr::sdk::ColorMultiply(color, colorationColor);
    memcpy(renderData->cbTriangle.color, &color, sizeof(gtf::scdr::sdk::Color));

    renderData->hasScissor = hasScissor(properties.scissor);
    if (renderData->hasScissor)
    {
        // write the scissor rectangle to a D3D11_RECT 
        renderData->scissor.left = properties.scissor[0];
        renderData->scissor.top = properties.scissor[1];
        renderData->scissor.right = properties.scissor[2];
        renderData->scissor.bottom = properties.scissor[3];
    }

    // Append the render data of this widget to the linked list.
    renderListAppend(renderData);

    return TRAVERSE_SUCCESS;
}

bool DirectX11TriangleRenderer::update(gtf::scdr::sdk::GraphicsContext&)
{
    // Here is typically code issued which updates graphics resources like
    // VBOs or textures. Graphics code for updating this resources would be issued 
    // with a GraphicsContext::executeUpdateFunction command.
    // Nothing to do for the triangle widget here.
    return true;
}

bool DirectX11TriangleRenderer::render(gtf::scdr::sdk::GraphicsContext& context_)
{
    // Walk over the linked list with the render data and issue the rendering for each object
    for (RenderData* renderData = renderListHead; renderData; renderData = renderData->next)
    {
        // Use the SCDR allocFrame to create a RenderFunctionContext structure. 
        RenderFunctionContext* renderFunctionContext = static_cast<RenderFunctionContext*>(context_.allocFrame(sizeof(RenderFunctionContext)));

        if (NULL == renderFunctionContext)
        {
            // Failed to allocate memory. Stop the rendering and return an error.
            return false;
        }

        // For rendering the per widget render data (RenderData) and the common data 
        // for all instances (UserData) are needed.
        renderFunctionContext->renderData = renderData;
        renderFunctionContext->userData = &userData;

        // Determine whether the widget is opaque.
        const bool isOpaque = renderData->cbTriangle.color[3] == 1.0f;

        // Generate a scheduler key. The SCDR ensures that all internal and external widgets
        // are rendered in a correct and fast order. The data in the schedule key is used for
        // sorting the render jobs.
        gtf::scdr::sdk::CScheduleDescriptor schedule(renderData->framebufferId, renderData->drawId, isOpaque);

        // Issue a the delete function doDelete for the RenderFunctionContext structure.
        // Use the executeDeleteFunction to ensure that the renderFunctionContext remains 
        // alive during rendering executed by executeRenderFunction.
        if (false == context_.executeDeleteFunction(doDelete, renderFunctionContext))
        {
            // In case the execute delete function was not issued successfully
            // free the memory in place and return an error.
            context_.freeFrame(renderFunctionContext);
            return false;
        }

        // Execute rendering by issuing the executeRenderFunction. 
        // Parameters are 
        // * the doRender function, which implements the rendering
        // * the renderFunctionContext structure with the render data
        // * the schedule structure with the schedule data.
        context_.executeRenderFunction(doRender, renderFunctionContext, schedule);
    }
    return true;
}

bool DirectX11TriangleRenderer::end(gtf::scdr::sdk::GraphicsContext&)
{
    // This method is called at the end of each frame rendered.
    // Reset the internal chain of custom widget data blocks. Since the memory blocks 
    // behind these pointers are managed by the SCDR there is no need for cleanup here.
    renderListHead = NULL;
    renderListTrail = NULL;
    return true;
}

bool DirectX11TriangleRenderer::doInit(void* userData_, gtf::scdr::sdk::GraphicsExecutionContext& context_)
{
    // Implementation of the initialization code of the renderer.
    // This method is issued by DirectX11TriangleRenderer::init. Since 
    // userData_ is a pointer to UserData, set in DirectX11TriangleRenderer::init,
    // the void pointer can be casted here.
    if (NULL == userData_)
    {
        return false;
    }

    UserData* data = static_cast<UserData*>(userData_);
    void* dev = NULL;

    // Get the D3DDevice from the SCDR render core.
    context_.getHandle(gtf::scdr::sdk::GraphicsHandles::D3D11Device, dev);
    ID3D11Device* d3dDevice = static_cast<ID3D11Device*>(dev);

    if (NULL == d3dDevice)
    {
        return false;
    }

    bool success = createShader(data, d3dDevice);
    success = (success) ? createVbo(data, d3dDevice) : false;
    success = (success) ? createCbo(data, d3dDevice) : false;
    success = (success) ? createDepthStencilState(data, d3dDevice) : false;
    success = (success) ? createRasterizerState(data, d3dDevice) : false;
    success = (success) ? createBlendState(data, d3dDevice) : false;

    if (false == success)
    {
        // If one init fails de-initialize everything and return false
        doDeinit(userData_, context_);
    }

    return success;
}

bool DirectX11TriangleRenderer::doRender(void* userData_, gtf::scdr::sdk::GraphicsExecutionContext& context_)
{
    // Implementation of the render code. This code was issued by GLES20TriangleRenderer::render
    // once for each widget to render.
    if (NULL == userData_)
    {
        return false;
    }

    // Cast the user data to the user specific data structures.
    RenderFunctionContext* renderFunctionContext = static_cast<RenderFunctionContext*>(userData_);
    UserData* userData = renderFunctionContext->userData;
    RenderData* renderData = renderFunctionContext->renderData;

    if (NULL == userData || NULL == renderData)
    {
        return false;
    }

    // Set the framebuffer to be used for this widget in the render context.
    context_.setFramebuffer(renderData->framebuffer);

    // Get the device context from the SCDR renderer
    void* dev = NULL;
    context_.getHandle(gtf::scdr::sdk::GraphicsHandles::D3D11DeviveContext, dev);
    ID3D11DeviceContext* d3dDeviceContext = static_cast<ID3D11DeviceContext*>(dev);

    // Check if all buffers are available
    bool success = checkUserData(userData);

    // Update the constant buffer with the matrices and the fill color
    success = (success) ? updateAndSetCbo(userData, renderData, d3dDeviceContext) : false;

    if (success)
    {
        ID3D11RenderTargetView* renderTargetView = NULL;
        ID3D11DepthStencilView* depthStencilView = NULL;

        // Get handles of render target view and depth stencil view from the SCDR renderer
        context_.getHandle(gtf::scdr::sdk::GraphicsHandles::D3D11RenderTargetView, renderData->framebuffer, reinterpret_cast<void*&>(renderTargetView));
        context_.getHandle(gtf::scdr::sdk::GraphicsHandles::D3D11DepthStencilView, renderData->framebuffer, reinterpret_cast<void*&>(depthStencilView));

        UINT stride = sizeof(float) * 2;
        UINT vboffset = 0;

        // Set the render state
        d3dDeviceContext->OMSetRenderTargets(1, &renderTargetView, depthStencilView);
        d3dDeviceContext->IASetVertexBuffers(0, 1, &(userData->vertexBuffer), &stride, &vboffset);
        d3dDeviceContext->IASetInputLayout(userData->inputLayout);
        d3dDeviceContext->IASetPrimitiveTopology(D3D11_PRIMITIVE_TOPOLOGY_TRIANGLELIST);
        d3dDeviceContext->VSSetShader(userData->vertexShader, 0, 0);
        d3dDeviceContext->PSSetShader(userData->pixelShader, 0, 0);

        updateViewport(renderData, context_, d3dDeviceContext);
        const bool hasAlpha = renderData->cbTriangle.color[3] != 1.0f;
        updatePipelineState(userData, d3dDeviceContext, hasAlpha, renderData->hasScissor);

        if (renderData->hasScissor)
        {
            d3dDeviceContext->RSSetScissorRects(1, &renderData->scissor);
        }

        d3dDeviceContext->Draw(3, 0);
    }

    return success;
}

bool DirectX11TriangleRenderer::doDeinit(void* userData_, gtf::scdr::sdk::GraphicsExecutionContext&)
{
    // Implementation of the deinit code. This code is issued by GLES20TriangleRenderer::deinit
    if (NULL == userData_)
    {
        return false;
    }

    UserData* userData = static_cast<UserData*>(userData_);

    if (userData->vertexShader) userData->vertexShader->Release();
    if (userData->pixelShader) userData->pixelShader->Release();
    if (userData->inputLayout) userData->inputLayout->Release();
    if (userData->vertexBuffer) userData->vertexBuffer->Release();
    if (userData->cbTriangleData) userData->cbTriangleData->Release();
    if (userData->depthStencisState) userData->depthStencisState->Release();
    if (userData->rasterizerState) userData->rasterizerState->Release();
    if (userData->rasterizerStateScissor) userData->rasterizerStateScissor->Release();
    if (userData->opaqueBlendState) userData->opaqueBlendState->Release();
    if (userData->transparentBlendState) userData->transparentBlendState->Release();

    memset(userData, 0, sizeof(UserData));
    return true;
}

bool DirectX11TriangleRenderer::doDelete(void* memory_, gtf::scdr::sdk::GraphicsExecutionContext& context_)
{
    // Implementation of the doDelete. This code is issued by GLES20TriangleRenderer::render to free the 
    // RenderFunctionContext structure
    context_.freeFrame(memory_);
    return true;
}

bool DirectX11TriangleRenderer::createDepthStencilState(UserData* userData_, ID3D11Device* d3dDevice_)
{
    D3D11_DEPTH_STENCIL_DESC depthStencilDesc;
    ZeroMemory(&depthStencilDesc, sizeof(depthStencilDesc));
    depthStencilDesc.DepthEnable = TRUE;
    depthStencilDesc.DepthWriteMask = D3D11_DEPTH_WRITE_MASK_ALL;
    depthStencilDesc.DepthFunc = D3D11_COMPARISON_LESS;
    depthStencilDesc.StencilEnable = FALSE;
    depthStencilDesc.StencilWriteMask = 0xff;
    depthStencilDesc.StencilReadMask = 0xff;
    depthStencilDesc.FrontFace.StencilFailOp = D3D11_STENCIL_OP_KEEP;
    depthStencilDesc.FrontFace.StencilDepthFailOp = D3D11_STENCIL_OP_KEEP;
    depthStencilDesc.FrontFace.StencilPassOp = D3D11_STENCIL_OP_KEEP;
    depthStencilDesc.FrontFace.StencilFunc = D3D11_COMPARISON_ALWAYS;
    depthStencilDesc.BackFace.StencilFailOp = D3D11_STENCIL_OP_KEEP;
    depthStencilDesc.BackFace.StencilDepthFailOp = D3D11_STENCIL_OP_KEEP;
    depthStencilDesc.BackFace.StencilPassOp = D3D11_STENCIL_OP_KEEP;
    depthStencilDesc.BackFace.StencilFunc = D3D11_COMPARISON_ALWAYS;

    const HRESULT hr = d3dDevice_->CreateDepthStencilState(&depthStencilDesc, &(userData_->depthStencisState));

    return SUCCEEDED(hr);
}

bool DirectX11TriangleRenderer::createRasterizerState(UserData* userData_, ID3D11Device* d3dDevice_)
{
    D3D11_RASTERIZER_DESC rasterizerDesc;
    ZeroMemory(&rasterizerDesc, sizeof(rasterizerDesc));
    rasterizerDesc.FillMode = D3D11_FILL_SOLID;
    rasterizerDesc.CullMode = D3D11_CULL_NONE;
    rasterizerDesc.FrontCounterClockwise = TRUE;
    rasterizerDesc.DepthBias = 0;
    rasterizerDesc.DepthBiasClamp = 0.0f;
    rasterizerDesc.SlopeScaledDepthBias = 0.0f;
    rasterizerDesc.DepthClipEnable = TRUE;
    rasterizerDesc.ScissorEnable = FALSE;
    rasterizerDesc.MultisampleEnable = FALSE;
    rasterizerDesc.AntialiasedLineEnable = FALSE;

    HRESULT hr = d3dDevice_->CreateRasterizerState(&rasterizerDesc, &(userData_->rasterizerState));

    if (FAILED(hr))
    {
        return false;
    }

    rasterizerDesc.ScissorEnable = TRUE;
    hr = d3dDevice_->CreateRasterizerState(&rasterizerDesc, &(userData_->rasterizerStateScissor));

    return SUCCEEDED(hr);
}

bool DirectX11TriangleRenderer::createBlendState(UserData* userData_, ID3D11Device* d3dDevice_)
{
    D3D11_RENDER_TARGET_BLEND_DESC renderTargetBlendDesc;
    ZeroMemory(&renderTargetBlendDesc, sizeof(renderTargetBlendDesc));
    renderTargetBlendDesc.BlendEnable = TRUE;
    renderTargetBlendDesc.BlendOp = D3D11_BLEND_OP_ADD;
    renderTargetBlendDesc.BlendOpAlpha = D3D11_BLEND_OP_ADD;
    renderTargetBlendDesc.SrcBlend = D3D11_BLEND_SRC_ALPHA;
    renderTargetBlendDesc.SrcBlendAlpha = D3D11_BLEND_ONE;
    renderTargetBlendDesc.DestBlend = D3D11_BLEND_INV_SRC_ALPHA;
    renderTargetBlendDesc.DestBlendAlpha = D3D11_BLEND_INV_SRC_ALPHA;
    renderTargetBlendDesc.RenderTargetWriteMask = 15;

    D3D11_BLEND_DESC blendDesc;
    ZeroMemory(&blendDesc, sizeof(blendDesc));
    blendDesc.AlphaToCoverageEnable = FALSE;
    blendDesc.IndependentBlendEnable = FALSE;

    for (int32_t i = 0; i < ARRAY_SIZE(blendDesc.RenderTarget); i++)
    {
        blendDesc.RenderTarget[i] = renderTargetBlendDesc;
    }

    HRESULT hr = d3dDevice_->CreateBlendState(&blendDesc, &(userData_->transparentBlendState));

    if (FAILED(hr))
    {
        return false;
    }

    renderTargetBlendDesc.BlendEnable = FALSE;
    renderTargetBlendDesc.BlendOp = D3D11_BLEND_OP_ADD;
    renderTargetBlendDesc.BlendOpAlpha = D3D11_BLEND_OP_ADD;
    renderTargetBlendDesc.SrcBlend = D3D11_BLEND_ONE;
    renderTargetBlendDesc.SrcBlendAlpha = D3D11_BLEND_ZERO;
    renderTargetBlendDesc.DestBlend = D3D11_BLEND_ZERO;
    renderTargetBlendDesc.DestBlendAlpha = D3D11_BLEND_ONE;

    for (int32_t i = 0; i < ARRAY_SIZE(blendDesc.RenderTarget); i++)
    {
        blendDesc.RenderTarget[i] = renderTargetBlendDesc;
    }

    hr = d3dDevice_->CreateBlendState(&blendDesc, &(userData_->opaqueBlendState));

    return SUCCEEDED(hr);
}

bool DirectX11TriangleRenderer::createCbo(UserData* userData_, ID3D11Device* d3dDevice_)
{
    D3D11_BUFFER_DESC cbBufDesc;
    ZeroMemory(&cbBufDesc, sizeof(cbBufDesc));
    cbBufDesc.ByteWidth = sizeof(ViewProjectionMaterialContstantBuffer);
    cbBufDesc.Usage = D3D11_USAGE_DYNAMIC;
    cbBufDesc.BindFlags = D3D11_BIND_CONSTANT_BUFFER;
    cbBufDesc.CPUAccessFlags = D3D11_CPU_ACCESS_WRITE;

    const HRESULT hr = d3dDevice_->CreateBuffer(&cbBufDesc, NULL, &(userData_->cbTriangleData));

    return SUCCEEDED(hr);
}

bool DirectX11TriangleRenderer::createVbo(UserData* userData_, ID3D11Device* d3dDevice_)
{
    static const float vertices[] =
    {
        // x-position, y-position, 3 vertices for the triangle
        0.5f, 1.0f,
        0.0f, 0.0f,
        1.0f, 0.0f
    };

    D3D11_BUFFER_DESC bufDesc;
    ZeroMemory(&bufDesc, sizeof(bufDesc));
    bufDesc.ByteWidth = sizeof(float) * _countof(vertices);
    bufDesc.Usage = D3D11_USAGE_IMMUTABLE;
    bufDesc.BindFlags = D3D11_BIND_VERTEX_BUFFER;

    D3D11_SUBRESOURCE_DATA bufData;
    ZeroMemory(&bufData, sizeof(bufData));
    bufData.pSysMem = vertices;

    const HRESULT hr = d3dDevice_->CreateBuffer(&bufDesc, &bufData, &(userData_->vertexBuffer));

    return SUCCEEDED(hr);
}

bool DirectX11TriangleRenderer::createShader(UserData* userData_, ID3D11Device* d3dDevice_)
{
    const char* msg = NULL;
    ID3DBlob* errorBlob = NULL;
    ID3DBlob* vertexShaderBlob = NULL;
    ID3DBlob* pixelShaderBlob = NULL;
    char vTarget[] = "vs_5_0";
    char pTarget[] = "ps_5_0";
    uint32_t flags = 0;

    // Vertex shader
    HRESULT hr = D3DCompile(VertexShader, strlen(VertexShader), NULL, NULL, NULL, "main", vTarget, flags, 0, &vertexShaderBlob, &errorBlob);
    if (FAILED(hr))
    {
        msg = static_cast<const char*>(errorBlob->GetBufferPointer());
        fprintf(stderr, "%s", msg);
        return false;
    }
    hr = d3dDevice_->CreateVertexShader(vertexShaderBlob->GetBufferPointer(), vertexShaderBlob->GetBufferSize(), 0, &(userData_->vertexShader));
    if (FAILED(hr))
    {
        return false;
    }

    // Pixel shader
    hr = D3DCompile(PixelShader, strlen(PixelShader), NULL, NULL, NULL, "main", pTarget, flags, 0, &(pixelShaderBlob), &errorBlob);
    if (FAILED(hr))
    {
        msg = static_cast<const char*>(errorBlob->GetBufferPointer());
        fprintf(stderr, "%s", msg);
        return false;
    }
    hr = d3dDevice_->CreatePixelShader(pixelShaderBlob->GetBufferPointer(), pixelShaderBlob->GetBufferSize(), 0, &(userData_->pixelShader));
    if (FAILED(hr))
    {
        return false;
    }

    // Input layout
    D3D11_INPUT_ELEMENT_DESC inputDesc[] =
    {
        { "POSITION", 0, DXGI_FORMAT_R32G32_FLOAT, 0, 0, D3D11_INPUT_PER_VERTEX_DATA, 0 }
    };

    uint32_t size = _countof(inputDesc);
    hr = d3dDevice_->CreateInputLayout(inputDesc, size, vertexShaderBlob->GetBufferPointer(), vertexShaderBlob->GetBufferSize(), &(userData_->inputLayout));

    if (NULL != errorBlob)
    {
        errorBlob->Release();
    }

    return true;
}

bool DirectX11TriangleRenderer::checkUserData(UserData* userData_)
{
    // Helper function to check if all buffers are available
    return (NULL != userData_) &&
           (NULL != userData_->vertexShader) &&
           (NULL != userData_->pixelShader) &&
           (NULL != userData_->inputLayout) &&
           (NULL != userData_->vertexBuffer) &&
           (NULL != userData_->cbTriangleData) &&
           (NULL != userData_->depthStencisState) &&
           (NULL != userData_->rasterizerState) &&
           (NULL != userData_->rasterizerStateScissor) &&
           (NULL != userData_->transparentBlendState) &&
           (NULL != userData_->opaqueBlendState);
}

void DirectX11TriangleRenderer::updatePipelineState(UserData* userData_, ID3D11DeviceContext* context_, bool hasAlpha_, bool hasScissor_)
{
    FLOAT blendFactor[] = { 0.0f, 0.0f, 0.0f, 0.0f };

    if (hasAlpha_)
    {
        context_->OMSetBlendState(userData_->transparentBlendState, blendFactor, 0xffffff);
    }
    else
    {
        context_->OMSetBlendState(userData_->opaqueBlendState, blendFactor, 0xffffff);
    }
    context_->OMSetDepthStencilState(userData_->depthStencisState, 0);

    if (hasScissor_)
    {
        context_->RSSetState(userData_->rasterizerStateScissor);
    }
    else
    {
        context_->RSSetState(userData_->rasterizerState);
    }
}

void DirectX11TriangleRenderer::updateViewport(RenderData* renderData_, gtf::scdr::sdk::GraphicsExecutionContext& context_, ID3D11DeviceContext* deviceContext_)
{
    D3D11_VIEWPORT vp;
    void* vpPtr = &vp;
    context_.getHandle(gtf::scdr::sdk::GraphicsHandles::Viewport, renderData_->framebuffer, vpPtr);
    deviceContext_->RSSetViewports(1, &vp);
}

bool DirectX11TriangleRenderer::updateAndSetCbo(UserData* userData_, RenderData* renderData_, ID3D11DeviceContext* context_)
{
    D3D11_MAPPED_SUBRESOURCE mapped;
    const UINT subresource = D3D11CalcSubresource(0, 0, 0);
    HRESULT hr = context_->Map(userData_->cbTriangleData, subresource, D3D11_MAP_WRITE_DISCARD, 0, &mapped);

    if (SUCCEEDED(hr))
    {
        memcpy(mapped.pData, &(renderData_->cbTriangle), sizeof(ViewProjectionMaterialContstantBuffer));
    }

    context_->Unmap(userData_->cbTriangleData, subresource);

    if (SUCCEEDED(hr))
    {
        context_->VSSetConstantBuffers(0, 1, &(userData_->cbTriangleData));
    }

    return SUCCEEDED(hr);
}

void DirectX11TriangleRenderer::renderListAppend(RenderData* data_)
{
    // Append data_ to the internal linked list with the user data.
    if (NULL == renderListHead)
    {
        renderListHead = data_;
        renderListTrail = data_;
    }
    else
    {
        renderListTrail->next = data_;
        renderListTrail = data_;
    }
}

bool DirectX11TriangleRenderer::hasScissor(const int16_t scissorRect_[])
{
    return (scissorRect_[0] != 0 ||
            scissorRect_[1] != 0 ||
            scissorRect_[2] != GtfLimits::int16_max ||
            scissorRect_[3] != GtfLimits::int16_max);
}