////////////////////////////////////////////////////////////////////////////////
// Copyright (c) Elektrobit Automotive GmbH
// Alle Rechte vorbehalten. All Rights Reserved.
//
// Information contained herein is subject to change without notice.
// Elektrobit retains ownership and all other rights in the software and each
// component thereof.
// Any reproduction of the software or components thereof without the prior
// written permission of Elektrobit is prohibited.
////////////////////////////////////////////////////////////////////////////////

#ifndef CUSTOM_WIDGET_GLES20_TRIANGLE_RENDERER_H_INCLUDED
#define CUSTOM_WIDGET_GLES20_TRIANGLE_RENDERER_H_INCLUDED

#include <gtf/scdr/sdk/WidgetRenderer.h>
#include <GtfUtils/math/matrix4x4.h>

#include <GtfUtils/smart_ptr/GtfRefCountedPtr.h>
#include <GtfResources/GtfImageResource.h>

#include <GLES2/gl2.h>

namespace gtf
{
namespace scdr
{
namespace sdk
{

class GraphicsContext;
class GraphicsExecutionContext;
class WidgetNode;
struct Framebuffer;

} // namespace sdk
} // namespace scdr
} // namespace gtf

//!
//! \brief This is the implementation of a custom renderer using the OpenGL|ES 2.0 API
//!
class GLES20TriangleRenderer : public gtf::scdr::sdk::WidgetRenderer
{
public:

    //!
    //! \brief Creates an instance of the custom renderer.
    //!
    //! \param[in] registration_ Interface used to get the handles for the widget properties from the SCDR.
    //! \param[in] context_ The SCDR graphics context.
    //! \param[in] userData_ User specific data.
    //!
    //! \return An instance of the custom renderer.
    //!
    //! \details This method is handed over to the SCDR at registration of the renderer at the 
    //!          SDK interface. Its called by the SCDR when the renderers are created.
    //!
    static gtf::scdr::sdk::WidgetRenderer* createIntance(
        const gtf::scdr::sdk::WidgetRendererRegistration& registration_,
        gtf::scdr::sdk::GraphicsContext& context_,
        void* userData_);

    //!
    //! \brief Destroys an instance of the custom renderer.
    //!
    //! \param[in] renderer_ The renderer to destroy.
    //! \param[in] registration_ Interface used to get the handles for the widget properties from the SCDR.
    //! \param[in] context_ The SCDR graphics context.
    //! \param[in] userData_ User specific data.
    //!
    //! \details This method is handed over to the SCDR at registration of the renderer at the 
    //!          SDK interface. Its called by the SCDR when the renderers are destroyed by the render core.
    //!
    static void destroyInstance(
        gtf::scdr::sdk::WidgetRenderer* renderer_,
        const gtf::scdr::sdk::WidgetRendererRegistration& registration_,
        gtf::scdr::sdk::GraphicsContext& context_,
        void* userData_);

    //!
    //! \brief Destructor
    //!
    virtual ~GLES20TriangleRenderer();

    //!
    //! \brief Initializes the renderer.
    //!
    //! \param[in] context_ The graphics context.
    //!
    //! \details Called once after instantiation of the renderer by the SCDR render core.
    //!
    virtual bool init(gtf::scdr::sdk::GraphicsContext& context_);

    //!
    //! \brief De-initializes the renderer.
    //!
    //! \param[in] context_ The graphics context.
    //!
    //! \details Called once before destruction of the renderer by the SCDR render core.
    //!
    virtual void deinit(gtf::scdr::sdk::GraphicsContext& context_);

    //!
    //! \brief Create method for a widget.
    //!
    //! \param[in] renderData_ A void pointer which can be used for custom per widget render data.
    //! \param[in] context_  Context for managing memory and graphics resources.
    //!
    //! \details Called after creating an instance of a widget by the SCDR render core.
    //!          If in the EB GUIDE Studio model a custom widget is instantiated three times, this method is called three times.
    //!
    virtual bool create(void* renderData_, gtf::scdr::sdk::GraphicsContext& context_);

    //!
    //! \brief Destroy method for a widget.
    //!
    //! \param[in] renderData_ A void pointer which can be used for custom per widget render data.
    //! \param[in] context_ Context for managing memory and graphics resources.
    //!
    //! \details Called before destroying an instance of a widget by the SCDR render core.
    //!          If in the guide model a custom widget is instantiated three times,
    //!          this method is called three times.
    //!
    virtual bool destroy(void* renderData_, gtf::scdr::sdk::GraphicsContext& context_);

    //!
    //! \brief Invoked by the renderer requesting graphics resources to be updated
    //!
    //! \param[in] context_ Context for managing memory and graphics resources.
    //!
    //! \return true if the update function succeeded, false otherwise.
    //!
    //! \details Called in every frame to render by the SCDR render core.
    //!
    virtual bool update(gtf::scdr::sdk::GraphicsContext& context_);

    //!
    //! \brief Invoked by the renderer requesting the widget to be drawn.
    //!
    //! \param[in] context_ Context for managing memory and graphics resources.
    //!
    //! \return true if the render function succeeded, false otherwise.
    //!
    //! \details Called in every frame to render by the SCDR render core.
    //!
    virtual bool render(gtf::scdr::sdk::GraphicsContext& context_);

    //!
    //! \brief Widget traversal function
    //!
    //! \param[in] renderData_       The widget's memory block for storing data required for drawing.
    //! \param[in] widget_           The widget being traversed.
    //! \param[in] traversalContext_ Context data required for traversing the widget.
    //! \param[in] graphicsContext_  Context for managing memory and graphics resources.
    //!
    //! \return The return code which may be (1) TAVERSE_SUCCESS, (2) TRAVERSE_DISCARD_WIDGET_AND_SUBTREE, or (3) TRAVERSER_APPEND_WIDGET_TO_EDIT_LIST
    //!
    virtual int32_t traverse(void* renderData_, gtf::scdr::sdk::Widget& widget_, gtf::scdr::sdk::TraverserContext& traverserContext_, gtf::scdr::sdk::GraphicsContext& graphicsContext_);

    //!
    //! \brief Invoked by the render core to notify the beginning of the frame.
    //!
    //! \param[in] context_ Context for managing memory and graphics resources.
    //!
    //! \return true if the function succeeded, false otherwise.
    //!
    virtual bool begin(gtf::scdr::sdk::GraphicsContext& context_);

    //!
    //! \brief Invoked by the render core to notify the end of the frame.
    //!
    //! \param[in] context_ Context for managing memory and graphics resources.
    //!
    //! \return true if the function succeeded, false otherwise.
    //!
    virtual bool end(gtf::scdr::sdk::GraphicsContext& context_);

private:

    // Custom Render functions
    static bool doInit(void* userData_, gtf::scdr::sdk::GraphicsExecutionContext& context_);
    static bool doRender(void* userData_, gtf::scdr::sdk::GraphicsExecutionContext& context_);
    static bool doDeinit(void* userData_, gtf::scdr::sdk::GraphicsExecutionContext& context_);
    static bool doDelete(void* memory_, gtf::scdr::sdk::GraphicsExecutionContext& context_);

    GLES20TriangleRenderer();

    struct ShaderParams
    {
        uint32_t                     program;
        uint32_t                     vertexShader;
        uint32_t                     fragmentShader;
        int32_t                      projMatLocation;
        int32_t                      modelMatLocation;
        int32_t                      colorLocation;
    };

    // Data used for all widget instances
    struct UserData
    {
        ShaderParams                 shader;
    } userData;

    // Data used for one specific widget instance.
    // This structure is mapped to the custom per widget memory
    // block provided by the SCDR. The size of this block is
    // configured in the WidgetRendererDescriptor, see
    // TrianglePlugin::registerRenderer
    struct RenderData
    {
        RenderData*                             next;
        gtf::scdr::sdk::WidgetDefaultProperties properties;
        float                                   fillColor[4];
        gtf::math::float4x4                     modelMat;
        uint16_t                                drawId;
    };

    // Helper structure to embrace UserData and RenderData
    struct RenderFunctionContext
    {
        UserData* userData;
        RenderData* renderData;
    };

    // Helper function to append a RenderData structure to the linked list.
    void renderListAppend(RenderData* data_);

    // Helper functions to init and load shader
    static bool initShader(ShaderParams& params_, const char* vertexShader_, const char* fragmentShader_);
    static GLuint loadShader(GLenum type_, const char* shaderSrc_);
    static bool hasScissor(const int16_t scissorRect_[]);
    static bool isOpaque(const float color_[]);

    // The handles for the custom properties. This might be an array if there is more than one custom handle.
    static uint16_t colorHandle;

    // Linked list for the custom widget data blocks.
    RenderData* renderListHead;
    RenderData* renderListTrail;

public:

    // The size of the custom render data, used by TrianglePlugin::registerRenderer
    static const uint32_t RenderDataSize = sizeof(RenderData);
};

#endif // CUSTOM_WIDGET_GLES20_TRIANGLE_RENDERER_H_INCLUDED